"""Flows to create and check training data."""

import logging
import traceback
from dataclasses import dataclass, field

from atomate2.forcefields.jobs import (
    ForceFieldRelaxMaker,
    ForceFieldStaticMaker,
)
from atomate2.vasp.jobs.base import BaseVaspMaker
from atomate2.vasp.jobs.core import StaticMaker
from atomate2.vasp.powerups import update_user_incar_settings
from atomate2.vasp.sets.core import StaticSetGenerator
from emmet.core.math import Matrix3D
from jobflow import Flow, Maker, Response, job
from pymatgen.core import Lattice
from pymatgen.core.structure import Structure
from pymatgen.io.ase import AseAtomsAdaptor

from autoplex.data.common.jobs import (
    convert_to_extxyz,
    generate_randomized_structures,
    get_supercell_job,
    plot_force_distribution,
)
from autoplex.data.common.utils import (
    ElementCollection,
    flatten,
)
from autoplex.misc.castep.jobs import CastepStaticMaker

__all__ = ["DFTStaticLabelling", "GenerateTrainingDataForTesting"]


logging.basicConfig(level=logging.DEBUG, format="[%(levelname)s] %(message)s")


@dataclass
class GenerateTrainingDataForTesting(Maker):
    """Maker for generating training data to test it and check the forces.

    This Maker will first generate training data based on the chosen ML model (default is GAP)
    by randomizing (ase rattle) atomic displacements in supercells of the provided input structures.
    Then it will proceed with MLIP-based Phonon calculations (based on atomate2 PhononMaker), collect
    all structure data in extended xyz files and plot the forces in histograms (per rescaling cell_factor
    and total).

    Parameters
    ----------
    name: str
        Name of the flow.
    bulk_relax_maker: ForceFieldRelaxMaker | None
        Maker for the relax jobs.
    static_energy_maker: ForceFieldStaticMaker | ForceFieldRelaxMaker | None
        Maker for the static jobs.

    """

    name: str = "generate_training_data_for_testing"
    bulk_relax_maker: ForceFieldRelaxMaker | None = None
    static_energy_maker: ForceFieldStaticMaker | ForceFieldRelaxMaker | None = None

    def make(
        self,
        train_structure_list: list[Structure],
        cell_factor_sequence: list[float] | None = None,
        potential_filename: str = "gap.xml",
        n_structures: int = 50,
        rattle_std: float = 0.01,
        relax_cell: bool = True,
        steps: int = 1000,
        supercell_matrix: Matrix3D | None = None,
        config_type: str = "train",
        x_min: int = 0,
        x_max: int = 5,
        bin_width: float = 0.125,
        **relax_kwargs,
    ):
        """
        Generate ase.rattled structures from the training data and returns histogram plots of the forces.

        Parameters
        ----------
        train_structure_list: list[Structure].
            List of pymatgen structures object.
        cell_factor_sequence: list[float]
            List of factor to resize cell parameters.
        potential_filename: str
            The param_file_name for :obj:`quippy.potential.Potential()'`.
        n_structures : int.
            Total number of randomly displaced structures to be generated.
        rattle_std: float.
            Rattle amplitude (standard deviation in normal distribution).
            Default=0.01.
        relax_cell : bool
            Whether to allow the cell shape/volume to change during relaxation.
        steps : int
            Maximum number of ionic steps allowed during relaxation.
        supercell_matrix: Matrix3D | None
            The matrix to generate the supercell.
        config_type: str
            Configuration type of the data.
        x_min: int
            Minimum value for the plot x-axis.
        x_max: int
            Maximum value for the plot x-axis.
        bin_width: float
            Width of the plot bins.
        relax_kwargs : dict
            Keyword arguments that will get passed to :obj:`Relaxer.relax`.

        Returns
        -------
        Matplotlib plots "count vs. forces".
        """
        jobs = []
        if cell_factor_sequence is None:
            cell_factor_sequence = [0.975, 1.0, 1.025, 1.05]
        for structure in train_structure_list:
            if self.bulk_relax_maker is None:
                self.bulk_relax_maker = ForceFieldRelaxMaker(
                    calculator_kwargs={
                        "args_str": "IP GAP",
                        "param_filename": str(potential_filename),
                    },
                    force_field_name="GAP",
                    relax_cell=relax_cell,
                    steps=steps,
                )
            if supercell_matrix is None:
                supercell_matrix = [[3, 0, 0], [0, 3, 0], [0, 0, 3]]

            bulk_relax = self.bulk_relax_maker.make(structure=structure)
            jobs.append(bulk_relax)
            supercell = get_supercell_job(
                structure=bulk_relax.output.structure,
                supercell_matrix=supercell_matrix,
            )
            jobs.append(supercell)

            for cell_factor in cell_factor_sequence:
                rattled_job = generate_randomized_structures(
                    structure=supercell.output,
                    n_structures=n_structures,
                    volume_custom_scale_factors=[cell_factor],
                    rattle_std=rattle_std,
                )
                jobs.append(rattled_job)
                static_conv_jobs = self.static_run_and_convert(
                    rattled_job.output,
                    cell_factor,
                    config_type,
                    potential_filename,
                    **relax_kwargs,
                )
                jobs.append(static_conv_jobs)
                plots = plot_force_distribution(
                    cell_factor, static_conv_jobs.output, x_min, x_max, bin_width
                )
                jobs.append(plots)

        return Flow(jobs=jobs, name=self.name)  # , plots.output)

    @job
    def static_run_and_convert(
        self,
        structure_list: list[Structure],
        cell_factor: float,
        config_type,
        potential_filename,
        **relax_kwargs,
    ):
        """
        Job for the static runs and the data conversion to the extxyz format.

        Parameters
        ----------
        structure_list: list[Structure].
            List of pymatgen structures object.
        cell_factor: float
            Factor to resize cell parameters.
        config_type: str
            Configuration type of the data.
        potential_filename: str
            The param_file_name for :obj:`quippy.potential.Potential()'`.
        relax_kwargs : dict
            Keyword arguments that will get passed to :obj:`Relaxer.relax`.

        """
        jobs = []
        for rattled in structure_list:
            if relax_kwargs == {}:
                relax_kwargs = {
                    "interval": 50000,
                    "fmax": 0.5,
                    "traj_file": rattled.reduced_formula
                    + "_"
                    + f"{cell_factor}".replace(".", "")
                    + ".pkl",
                }
            if self.static_energy_maker is None:
                self.static_energy_maker = ForceFieldRelaxMaker(
                    calculator_kwargs={
                        "args_str": "IP GAP",
                        "param_filename": str(potential_filename),
                    },
                    force_field_name="GAP",
                    relax_cell=False,
                    relax_kwargs=relax_kwargs,
                    steps=1,
                )
            static_run = self.static_energy_maker.make(structure=rattled)
            jobs.append(static_run)
            conv_job = convert_to_extxyz(
                static_run.output,
                rattled.reduced_formula
                + "_"
                + f"{cell_factor}".replace(".", "")
                + ".pkl",
                config_type,
                f"{cell_factor}".replace(".", ""),
            )
            jobs.append(conv_job)

        return Response(replace=Flow(jobs), output=conv_job.output)


@dataclass
class DFTStaticLabelling(Maker):
    """
    Maker to set up and run VASP static calculations for input structures, including bulk, isolated atoms, and dimers.

    It supports custom VASP input parameters and error handlers.

    Parameters
    ----------
    name: str
        Name of the flow.
    isolated_atom: bool
        If true, perform single-point calculations for isolated atoms. Default is False.
    isolated_species: list[str] | None
        List of species for which to perform isolated atom calculations. If None,
        species will be automatically derived from the 'structures' list. Default is None.
    e0_spin: bool
        If true, include spin polarization in isolated atom and dimer calculations.
        Default is False.
    isolatedatom_box: list[float]
        List of the lattice constants for a isolated_atom configuration.
    dimer: bool
        If true, perform single-point calculations for dimers. Default is False.
    dimer_box: list[float]
        The lattice constants of a dimer box.
    dimer_species: list[str] | None
        List of species for which to perform dimer calculations. If None, species
        will be derived from the 'structures' list. Default is None.
    dimer_range: list[float] | None
        Range of distances for dimer calculations.
    dimer_num: int
        Number of different distances to consider for dimer calculations.
    custom_incar: dict | None
        Dictionary of custom VASP input parameters. If provided, will update the
        default parameters. Default is None.
    custom_potcar: dict | None
        Dictionary of POTCAR settings to update. Keys are element symbols, values are the desired POTCAR labels.
        Default is None.
    static_energy_maker: BaseVaspMaker | CastepStaticMaker | ForceFieldStaticMaker
        Maker for static energy jobs: either BaseVaspMaker (VASP-based) or CastepStaticMaker (CASTEP-based) or
        ForceFieldStaticMaker (force field-based). Defaults to StaticMaker (VASP-based).
    static_energy_maker_isolated_atoms: BaseVaspMaker | ForceFieldStaticMaker | None
        Maker for static energy jobs of isolated atoms: either BaseVaspMaker (VASP-based) or
        ForceFieldStaticMaker (force field-based) or None. If set to `None`, the parameters
        from `static_energy_maker` will be used as the default for isolated atoms. In this case,
        if `static_energy_maker` is a `StaticMaker`, all major settings will be inherited,
        except that `kspacing` will be automatically set to 100 to enforce a Gamma-point-only calculation.
        This is typically suitable for single-atom systems. Default is None. If a non-`StaticMaker` maker
        is used here, its output must include a `dir_name` field to ensure compatibility with downstream workflows.

    Returns
    -------
    dict
        A dictionary containing:
        - 'dirs_of_dft': List of directories containing DFT data.
        - 'config_type': List of configuration types corresponding to each directory.
    """

    name: str = "do_dft_labelling"
    isolated_atom: bool = False
    isolated_species: list[str] | None = None
    e0_spin: bool = False
    isolatedatom_box: list[float] = field(default_factory=lambda: [20, 20, 20])
    dimer: bool = False
    dimer_box: list[float] = field(default_factory=lambda: [20, 20, 20])
    dimer_species: list[str] | None = None
    dimer_range: list[float] | None = None
    dimer_num: int = 21
    custom_incar: dict | None = None
    custom_potcar: dict | None = None
    static_energy_maker: BaseVaspMaker | CastepStaticMaker | ForceFieldStaticMaker = (
        field(
            default_factory=lambda: StaticMaker(
                input_set_generator=StaticSetGenerator(
                    user_incar_settings={
                        "ADDGRID": "True",
                        "ENCUT": 520,
                        "EDIFF": 1e-06,
                        "ISMEAR": 0,
                        "SIGMA": 0.01,
                        "PREC": "Accurate",
                        "ISYM": None,
                        "KSPACING": 0.2,
                        "NPAR": 8,
                        "LWAVE": "False",
                        "LCHARG": "False",
                        "ENAUG": None,
                        "GGA": None,
                        "ISPIN": None,
                        "LAECHG": None,
                        "LELF": None,
                        "LORBIT": None,
                        "LVTOT": None,
                        "NSW": None,
                        "SYMPREC": None,
                        "NELM": 100,
                        "LMAXMIX": None,
                        "LASPH": None,
                        "AMIN": None,
                    }
                ),
                run_vasp_kwargs={"handlers": ()},
            )
        )
    )
    static_energy_maker_isolated_atoms: (
        BaseVaspMaker | CastepStaticMaker | ForceFieldStaticMaker | None
    ) = None

    @job
    def make(
        self,
        structures: list,
        config_type: str | None = None,
    ):
        """
        Maker to set up and run VASP static calculations.

        Parameters
        ----------
        structures : list[Structure] | list[list[Structure]]
            List of structures for which to run the VASP static calculations. If None,
            no bulk calculations will be performed. Default is None.
        config_type : str
            Configuration types corresponding to the structures. If None, defaults
            to 'bulk'. Default is None.
        """
        job_list = []

        if isinstance(structures[0], list):
            structures = flatten(structures, recursive=False)

        dirs: dict[str, list[str]] = {"dirs_of_dft": [], "config_type": []}

        if isinstance(self.static_energy_maker, StaticMaker):

            if self.custom_incar is not None:
                self.static_energy_maker.input_set_generator.user_incar_settings.update(
                    self.custom_incar
                )

            if self.custom_potcar is not None:
                self.static_energy_maker.input_set_generator.user_potcar_settings.update(
                    self.custom_potcar
                )

        st_m = self.static_energy_maker

        if structures:
            for idx, struct in enumerate(structures):
                static_job = st_m.make(structure=struct)
                static_job.name = f"static_bulk_{idx}"
                dirs["dirs_of_dft"].append(static_job.output.dir_name)
                if config_type:
                    dirs["config_type"].append(config_type)
                else:
                    dirs["config_type"].append("bulk")
                job_list.append(static_job)

        if self.isolated_atom:
            try:
                if self.isolated_species is not None:
                    syms = self.isolated_species

                elif (self.isolated_species is None) and (structures is not None):
                    # Get the species from the database
                    atoms = [AseAtomsAdaptor().get_atoms(at) for at in structures]
                    syms = ElementCollection(atoms).get_species()

                for idx, sym in enumerate(syms):
                    lattice = Lattice.orthorhombic(
                        self.isolatedatom_box[0],
                        self.isolatedatom_box[1],
                        self.isolatedatom_box[2],
                    )
                    isolated_atom_struct = Structure(lattice, [sym], [[0.0, 0.0, 0.0]])

                    if self.static_energy_maker_isolated_atoms is None:
                        static_job = st_m.make(structure=isolated_atom_struct)
                        if isinstance(self.static_energy_maker, StaticMaker):
                            static_job = update_user_incar_settings(
                                static_job,
                                {"KSPACING": 100.0, "KPAR": 1},
                            )

                            if self.e0_spin:
                                static_job = update_user_incar_settings(
                                    static_job, {"ISPIN": 2}
                                )
                    else:
                        static_job = self.static_energy_maker_isolated_atoms.make(
                            structure=isolated_atom_struct
                        )

                    static_job.name = f"static_isolated_{idx}"
                    dirs["dirs_of_dft"].append(static_job.output.dir_name)
                    dirs["config_type"].append("IsolatedAtom")
                    job_list.append(static_job)

            except ValueError as e:
                logging.error(f"Unknown species of isolated atoms! Exception: {e}")
                traceback.print_exc()

        if self.dimer:
            try:
                atoms = [AseAtomsAdaptor().get_atoms(at) for at in structures]
                if self.dimer_species is not None:
                    dimer_syms = self.dimer_species
                elif (self.dimer_species is None) and (structures is not None):
                    # Get the species from the database
                    dimer_syms = ElementCollection(atoms).get_species()
                pairs_list = ElementCollection(atoms).find_element_pairs(dimer_syms)
                for pair in pairs_list:
                    for dimer_i in range(self.dimer_num):
                        if self.dimer_range is not None:
                            dimer_distance = self.dimer_range[0] + (
                                self.dimer_range[1] - self.dimer_range[0]
                            ) * float(dimer_i) / float(
                                self.dimer_num - 1 + 0.000000000001
                            )

                        lattice = Lattice.orthorhombic(
                            self.dimer_box[0],
                            self.dimer_box[1],
                            self.dimer_box[2],
                        )
                        dimer_struct = Structure(
                            lattice,
                            [pair[0], pair[1]],
                            [[0.0, 0.0, 0.0], [dimer_distance, 0.0, 0.0]],
                            coords_are_cartesian=True,
                        )

                        static_energy_maker_dimer = (
                            self.static_energy_maker_isolated_atoms
                        )
                        if static_energy_maker_dimer is None:
                            static_job = st_m.make(structure=dimer_struct)
                            if isinstance(self.static_energy_maker, StaticMaker):
                                static_job = update_user_incar_settings(
                                    static_job,
                                    {"KSPACING": 100.0, "KPAR": 1},
                                )

                                if self.e0_spin:
                                    static_job = update_user_incar_settings(
                                        static_job, {"ISPIN": 2}
                                    )
                        else:
                            static_job = static_energy_maker_dimer.make(
                                structure=dimer_struct
                            )

                        static_job.name = f"static_dimer_{dimer_i}"
                        dirs["dirs_of_dft"].append(static_job.output.dir_name)
                        dirs["config_type"].append("dimer")
                        job_list.append(static_job)

            except ValueError:
                logging.error("Unknown atom types in dimers!")
                traceback.print_exc()

        return Response(replace=Flow(job_list), output=dirs)



"""Document Schema for CASTEP.

The following code has been taken and modified from
https://github.com/materialsproject/atomate2/blob/main/src/atomate2/ase/schemas.py

The code has been released under BSD 3-Clause License
and the following copyright applies:
atomate2 Copyright (c) 2015, The Regents of the University of
California, through Lawrence Berkeley National Laboratory (subject
to receipt of any required approvals from the U.S. Dept. of Energy).
All rights reserved.

The original schema definitions there haven been taken and generalized to
generic ASE calculators from
https://github.com/materialsvirtuallab/m3gnet
The code has been released under BSD 3-Clause License
and the following copyright applies:
Copyright (c) 2022, Materials Virtual Lab.
"""

from emmet.core.math import Matrix3D, Vector3D
from emmet.core.structure import StructureMetadata
from pydantic import BaseModel, Field
from pymatgen.core.structure import Structure


class InputDoc(BaseModel):
    """The inputs used to run this job."""

    input_set: dict | None = Field(
        None, description="Input set describing the input for CASTEP."
    )


class OutputDoc(BaseModel):
    """The outputs of this job."""

    structure: Structure | None = Field(
        None, description="Final output structure from the task."
    )

    energy: float | None = Field(None, description="Total energy in units of eV.")

    energy_per_atom: float | None = Field(
        None,
        description="Energy per atom of the final molecule or structure "
        "in units of eV/atom.",
    )

    forces: list[Vector3D] | None = Field(
        None,
        description=(
            "The force on each atom in units of eV/A for the final molecule "
            "or structure."
        ),
    )

    # NOTE: units for stresses were converted to kbar (* -10 from standard output)
    #       to comply with MP convention
    stress: Matrix3D | None = Field(
        None, description="The stress on the cell in units of kbar."
    )


class TaskDoc(StructureMetadata):
    """Document containing information on structure manipulation using CASTEP."""

    structure: Structure = Field(
        None, description="Final output structure from the task"
    )

    input: InputDoc = Field(
        None, description="The input information used to run this job."
    )

    output: OutputDoc = Field(None, description="The output information from this job.")

    task_label: str = Field(
        None,
        description="Description of the CASTEP task (e.g., static, relax)",
    )

    dir_name: str | None = Field(
        None, description="Directory where the ASE calculations are performed."
    )



"""Utility functions for fitting jobs."""

import contextlib
import logging
import multiprocessing as mp
import os
import re
import shutil
import subprocess
import sys
import xml.etree.ElementTree as ET
from collections.abc import Iterable
from functools import partial
from pathlib import Path
from typing import Literal

import ase
import lightning as pl
import matgl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import quippy.potential
import torch
from ase.atoms import Atoms
from ase.calculators.singlepoint import SinglePointCalculator
from ase.constraints import voigt_6_to_full_3x3_stress
from ase.data import chemical_symbols
from ase.io import read, write
from ase.io.extxyz import XYZError
from atomate2.utils.path import strip_hostname
from calorine.nep import read_loss, write_nepfile, write_structures
from dgl.data.utils import split_dataset
from matgl.apps.pes import Potential
from matgl.ext.pymatgen import Structure2Graph, get_element_list
from matgl.graph.data import MGLDataLoader, MGLDataset, collate_fn_pes
from matgl.models import M3GNet
from matgl.utils.training import PotentialLightningModule
from monty.dev import requires
from monty.serialization import dumpfn
from nequip.ase import NequIPCalculator
from numpy import ndarray
from pydantic import Field
from pymatgen.core import Structure
from pymatgen.io.ase import AseAtomsAdaptor
from pytorch_lightning.loggers import CSVLogger
from quippy import descriptors
from scipy.spatial import ConvexHull
from threadpoolctl import threadpool_limits

from autoplex import (
    GAP_HYPERS,
    JACE_HYPERS,
    M3GNET_HYPERS,
    MACE_HYPERS,
    NEP_HYPERS,
    NEQUIP_HYPERS,
)
from autoplex.data.common.utils import (
    data_distillation,
    plot_energy_forces,
    rms_dict,
    stratified_dataset_split,
)

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)


def gap_fitting(
    db_dir: Path,
    species_list: list | None = None,
    hyperparameters: GAP_HYPERS = GAP_HYPERS,
    num_processes_fit: int = 32,
    auto_delta: bool = True,
    glue_xml: bool = False,
    ref_energy_name: str = "REF_energy",
    ref_force_name: str = "REF_forces",
    ref_virial_name: str = "REF_virial",
    train_name: str = "train.extxyz",
    test_name: str = "test.extxyz",
    glue_file_path: str = "glue.xml",
    fit_kwargs: dict | None = None,  # pylint: disable=E3701
) -> dict:
    """
    Perform the GAP (Gaussian approximation potential) model fitting.

    Parameters
    ----------
    db_dir: str or path
        Path to database directory.
    species_list: list
        List of element names (strings)
    hyperparameters: MLIP_HYPERS.GAP
        Fit hyperparameters.
    num_processes_fit: int
        Number of processes used for gap_fit
    auto_delta: bool
        Automatically determine delta for 2b, 3b and soap terms.
    glue_xml: bool
        Use the glue.xml core potential instead of fitting 2b terms.
    ref_energy_name: str
        Reference energy name.
    ref_force_name : str
        Reference force name.
    ref_virial_name: str
        Reference virial name.
    train_name: str
        Name of the training set file.
    test_name: str
        Name of the test set file.
    glue_file_path: str
        Name of the glue.xml file path.
    fit_kwargs: dict
        Additional keyword arguments for GAP fitting with keys same as
        those in gap-defaults.json.

    Returns
    -------
    dict
        A dictionary with train_error, test_error, path_to_mlip

    """
    hyperparameters = hyperparameters.model_copy(deep=True)
    # keep additional pre- and suffixes
    gap_file_xml = train_name.replace("train", "gap_file").replace(".extxyz", ".xml")
    quip_train_file = train_name.replace("train", "quip_train")
    quip_test_file = test_name.replace("test", "quip_test")
    mlip_path: Path = prepare_fit_environment(
        db_dir, Path.cwd(), glue_xml, train_name, test_name, glue_file_path
    )

    db_atoms = ase.io.read(os.path.join(db_dir, train_name), index=":")
    train_data_path = os.path.join(db_dir, train_name)

    test_data_path = os.path.join(db_dir, test_name)

    hyperparameters.update_parameters(
        {
            "general": {
                "gp_file": gap_file_xml,
                "energy_parameter_name": ref_energy_name,
                "force_parameter_name": ref_force_name,
                "virial_parameter_name": ref_virial_name,
            }
        }
    )

    if fit_kwargs:
        hyperparameters.update_parameters(fit_kwargs)

    gap_default_hyperparameters = hyperparameters.model_dump(by_alias=True)

    include_two_body = gap_default_hyperparameters["general"]["two_body"]
    include_three_body = gap_default_hyperparameters["general"]["three_body"]
    include_soap = gap_default_hyperparameters["general"]["soap"]
    gap_default_hyperparameters["general"].update({"at_file": train_data_path})

    if auto_delta:
        if include_two_body and not glue_xml:
            if include_three_body or include_soap:
                cutoff_2b = gap_default_hyperparameters["twob"]["cutoff"]
                delta_2b = calculate_delta_2b(
                    atoms_db=db_atoms, cutoff=cutoff_2b, e_name=ref_energy_name
                )
            else:
                delta_2b = 1

            gap_default_hyperparameters["twob"].update({"delta": delta_2b})

            fit_parameters_list = gap_hyperparameter_constructor(
                gap_parameter_dict=gap_default_hyperparameters,
                include_two_body=include_two_body,
            )

            run_gap(num_processes_fit, fit_parameters_list)

            run_ase_gap(
                num_processes_fit, train_data_path, gap_file_xml, quip_train_file
            )

        if include_three_body and not glue_xml:
            if include_two_body:
                cutoff_3b = gap_default_hyperparameters["threeb"]["cutoff"]
                energy_residual = energy_remain(quip_train_file)
                descriptor_num_list = [
                    compute_num_of_descriptor(atom=at, nb=3, cutoff=cutoff_3b) / len(at)
                    for at in db_atoms
                ]
                num_of_descriptors = sum(descriptor_num_list) / len(db_atoms)
                delta_3b = energy_residual / num_of_descriptors
            else:
                delta_3b = 1
            gap_default_hyperparameters["threeb"].update({"delta": delta_3b})

            fit_parameters_list = gap_hyperparameter_constructor(
                gap_parameter_dict=gap_default_hyperparameters,
                include_two_body=include_two_body,
                include_three_body=include_three_body,
            )

            print("fit_parameters_list:", fit_parameters_list)

            run_gap(num_processes_fit, fit_parameters_list)
            run_ase_gap(
                num_processes_fit, train_data_path, gap_file_xml, quip_train_file
            )

        if glue_xml:
            gap_default_hyperparameters["general"].update(
                {"core_param_file": "glue.xml"}
            )
            gap_default_hyperparameters["general"].update({"core_ip_args": "{IP Glue}"})

            fit_parameters_list = gap_hyperparameter_constructor(
                gap_parameter_dict=gap_default_hyperparameters,
                include_two_body=False,
                include_three_body=False,
            )

            run_gap(num_processes_fit, fit_parameters_list)
            run_ase_gap(
                num_processes_fit,
                train_data_path,
                gap_file_xml,
                quip_train_file,
                glue_xml,
            )

        if include_soap:
            delta_soap = (
                energy_remain(quip_train_file)
                if include_two_body or include_three_body
                else 1
            )
            gap_default_hyperparameters["soap"].update({"delta": delta_soap})

            fit_parameters_list = gap_hyperparameter_constructor(
                gap_parameter_dict=gap_default_hyperparameters,
                include_two_body=include_two_body,
                include_three_body=include_three_body,
                include_soap=include_soap,
            )

            run_gap(num_processes_fit, fit_parameters_list)
            run_ase_gap(
                num_processes_fit,
                train_data_path,
                gap_file_xml,
                quip_train_file,
                glue_xml,
            )

    else:
        if glue_xml:
            gap_default_hyperparameters["general"].update(
                {"core_param_file": "glue.xml"}
            )
            gap_default_hyperparameters["general"].update({"core_ip_args": "{IP Glue}"})

        fit_parameters_list = gap_hyperparameter_constructor(
            gap_parameter_dict=gap_default_hyperparameters,
            include_two_body=include_two_body,
            include_three_body=include_three_body,
            include_soap=include_soap,
        )

        run_gap(num_processes_fit, fit_parameters_list)
        run_ase_gap(
            num_processes_fit,
            train_data_path,
            gap_file_xml,
            quip_train_file,
            glue_xml,
        )

    # Calculate training error

    train_error = energy_remain(quip_train_file)
    logging.info(f"Training error of MLIP (eV/at.): {round(train_error, 7)}")

    # Calculate testing error
    run_ase_gap(
        num_processes_fit, test_data_path, gap_file_xml, quip_test_file, glue_xml
    )
    test_error = energy_remain(quip_test_file)
    logging.info(f"Testing error of MLIP (eV/at.): {round(test_error, 7)}")

    if not glue_xml and species_list:
        try:
            plot_energy_forces(
                title="Data error metrics",
                energy_limit=0.005,
                force_limit=0.1,
                species_list=species_list,
                train_name=train_name,
                test_name=test_name,
            )
        except (ValueError, XYZError) as e:
            logging.warning(f"Skipped fit error metrics plot because of: \n{e}")

    return {
        "train_error": train_error,
        "test_error": test_error,
        "mlip_path": mlip_path,
    }


@requires(
    (
        subprocess.run(
            'julia -e "using Pkg; println(haskey(Pkg.dependencies(), '
            'Base.UUID(\\"3b96b61c-0fcc-4693-95ed-1ef9f35fcc53\\")))"',
            shell=True,
            capture_output=True,
            text=True,
            check=False,
        ).stdout.strip()
    )
    == "true",
    "J-ACE fitting requires the executable 'julia' and ACEPotentials.jl v0.6.7 library to be in PATH. "
    "Please follow the instructions in the autoplex documentation to install the required julia dependencies "
    "and add them to PATH. Link to the documentation:"
    " https://autoatml.github.io/autoplex/user/index.html#standard-installation",
)
def jace_fitting(
    db_dir: str | Path,
    hyperparameters: JACE_HYPERS = JACE_HYPERS,
    isolated_atom_energies: dict | None = None,
    ref_energy_name: str = "REF_energy",
    ref_force_name: str = "REF_forces",
    ref_virial_name: str = "REF_virial",
    num_processes_fit: int = 32,
    fit_kwargs: dict | None = None,
) -> dict:
    """
    Perform the ACE (Atomic Cluster Expansion) potential fitting.

    This function sets up and executes a Julia script to perform ACE fitting using specified parameters
    and input data located in the provided directory. It handles the input/output of atomic configurations,
    sets up the ACE model, and calculates training and testing errors after fitting.

    Parameters
    ----------
    db_dir: str or Path
        directory containing the training and testing data files.
    hyperparameters: MLIP_HYPERS.J_ACE
        Fit hyperparameters.
    isolated_atom_energies: dict:
        mandatory dictionary mapping element numbers to isolated energies.
    ref_energy_name : str, optional
        Reference energy name.
    ref_force_name : str, optional
        Reference force name.
    ref_virial_name : str, optional
        Reference virial name.
    num_processes_fit: int
        number of processes to use for parallel computation.
    fit_kwargs: dict.
        optional dictionary with parameters for ace fitting with keys same as
        mlip-rss-defaults.json.

    Keyword Arguments
    -----------------
    order: int
        order of ACE.
    totaldegree: int
        total degree of the polynomial terms in the ACE model.
    cutoff: float
        cutoff distance for atomic interactions in the ACE model.
    solver: str
        solver to be used for fitting the ACE model. Default is "BLR" (Bayesian Linear Regression).
        For very large-scale parameter estimation problems, using "LSQR" solver.

    Returns
    -------
    dict[str, float]
        A dictionary containing train_error, test_error, and the path to the fitted MLIP.

    Raises
    ------
    - ValueError: If the `isolated_atom_energies` dictionary is empty or not provided when required.
    """
    hyperparameters = hyperparameters.model_copy(deep=True)
    train_atoms = ase.io.read(os.path.join(db_dir, "train.extxyz"), index=":")
    source_file_path = os.path.join(db_dir, "test.extxyz")
    shutil.copy(source_file_path, ".")
    isolated_atom_energies_update = {}

    if isolated_atom_energies:
        for e_num, e_energy in isolated_atom_energies.items():
            isolated_atom_energies_update[chemical_symbols[int(e_num)]] = e_energy
    else:
        raise ValueError("isolated_atom_energies parameter is empty or not defined!")

    formatted_isolated_atom_energies = (
        "["
        + ", ".join(
            [
                f":{key} => {value}"
                for key, value in isolated_atom_energies_update.items()
            ]
        )
        + "]"
    )
    formatted_species = (
        "["
        + ", ".join([f":{key}" for key, value in isolated_atom_energies_update.items()])
        + "]"
    )

    train_ace = [
        at for at in train_atoms if "IsolatedAtom" not in at.info["config_type"]
    ]
    ase.io.write("train_ace.extxyz", train_ace, format="extxyz")

    if fit_kwargs:
        hyperparameters.update_parameters(fit_kwargs)

    jace_hypers = hyperparameters.model_dump(by_alias=True)

    order = jace_hypers["order"]
    totaldegree = jace_hypers["totaldegree"]
    cutoff = jace_hypers["cutoff"]
    solver = jace_hypers["solver"]

    ace_text = f"""using ACEpotentials
using LinearAlgebra: norm, Diagonal
using CSV, DataFrames
using Distributed
addprocs({num_processes_fit - 1}, exeflags="--project=$(Base.active_project())")
@everywhere using ACEpotentials

data_file = "train_ace.extxyz"
data = read_extxyz(data_file)
test_data_file = "test.extxyz"
test_data = read_extxyz(test_data_file)
data_keys = (energy_key = "{ref_energy_name}", force_key = "{ref_force_name}", virial_key = "{ref_virial_name}")

model = acemodel(elements={formatted_species},
                order={order},
                totaldegree={totaldegree},
                rcut={cutoff},
                Eref={formatted_isolated_atom_energies})

weights = Dict(
            "crystal" => Dict("E" => 30.0, "F" => 1.0 , "V" => 1.0 ),
            "RSS" => Dict("E" => 3.0, "F" => 0.5 , "V" => 0.1 ),
            "amorphous" => Dict("E" => 3.0, "F" => 0.5 , "V" => 0.1 ),
            "liquid" => Dict("E" => 10.0, "F" => 0.5 , "V" => 0.25 ),
            "RSS_initial" => Dict("E" => 1.0, "F" => 0.5 , "V" => 0.1 ),
            "dimer" => Dict("E" => 30.0, "F" => 1.0 , "V" => 1.0 )
            )

P = smoothness_prior(model; p = 4)

solver = ACEfit.{solver}()

acefit!(model, data; solver=solver, weights=weights, prior = P, data_keys...)

@info("Training Error Table")
ACEpotentials.linear_errors(data, model; data_keys...)

@info("Testing Error Table")
ACEpotentials.linear_errors(test_data, model; data_keys...)

@info("Manual RMSE Test")
potential = model.potential
train_energies = [ JuLIP.get_data(at, "{ref_energy_name}") / length(at) for at in data]
model_energies_train = [energy(potential, at) / length(at) for at in data]
rmse_energy_train = norm(train_energies - model_energies_train) / sqrt(length(data))
test_energies = [ JuLIP.get_data(at, "{ref_energy_name}") / length(at) for at in test_data]
model_energies_pred = [energy(potential, at) / length(at) for at in test_data]
rmse_energy_test = norm(test_energies - model_energies_pred) / sqrt(length(test_data))

df = DataFrame(rmse_energy_train = rmse_energy_train, rmse_energy_test = rmse_energy_test)
CSV.write("rmse_energies.csv", df)

save_potential("acemodel.json", model)
export2lammps("acemodel.yace", model)
    """

    with open("ace.jl", "w") as file:
        file.write(ace_text)

    os.system(f"export OMP_NUM_THREADS={num_processes_fit} && julia ace.jl")

    energy_err = pd.read_csv("rmse_energies.csv")
    train_error = energy_err["rmse_energy_train"][0]
    test_error = energy_err["rmse_energy_test"][0]

    return {
        "train_error": train_error,
        "test_error": test_error,
        "mlip_path": Path.cwd(),
    }


def nep_fitting(
    db_dir: str | Path,
    hyperparameters: NEP_HYPERS = NEP_HYPERS,
    ref_energy_name: str = "REF_energy",
    ref_force_name: str = "REF_forces",
    ref_virial_name: str = "REF_virial",
    species_list: list | None = None,
    gpu_identifier_indices: list[int] = list[0],
    fit_kwargs: dict | None = None,
) -> dict:
    """
    Perform the NEP (Neural evolution Potential) model fitting.

    Parameters
    ----------
    db_dir: Path
        Directory containing the training and testing data files.
    path_to_hyperparameters : str or Path.
        Path to JSON file containing the M3GNet hyperparameters.
    ref_energy_name : str, optional
        Reference energy name.
    ref_force_name : str, optional
        Reference force name.
    ref_virial_name : str, optional
        Reference virial name.
    species_list: list
        List of element names (strings)
    gpu_identifier_indices: list[int]
        Indices that identifies the GPU that NEP should be run with
    fit_kwargs: dict.
        optional dictionary with parameters for NEP fitting with keys same as
        mlip-rss-defaults.json.

    Keyword Arguments
    -----------------
    version: int
        NEP model version to train can be 3 or 4. Default is 4.
    type: list[int, str]
        Number of atom types and list of chemical species. Number
        of atom types must be an integer, followed by chemical
        symbols of species as in periodic table for which model
        needs to be trained, separated by comma.
        Default is [1, "X"] as a placeholder. Example:
        [2, "Pb", "Te"].
    type_weight: float
        Weights for different chemical species. Default is 1.0
    model_type: int
        Type of model that is being trained. Can be 0 (potential),
        1 (dipole), 2 (polarizability). Default is 0.
    prediction: int
        Mode of NEP run. Set 0 for training and 1 for inference.
        Default is 0.
    cutoff: list[int, int]
        Radial and angular cutoff. First element is for radial cutoff and
        second element is for angular cutoff. Default is [6, 5].
    n_max: list[int, int]
        Number of radial and angular descriptors. First element is for radial
        and second element is for angular. Default is [4, 4].
    basis_size: list[int, int]
        Number of basis functions that are used to build the radial and angular
        descriptor. First element is for radial descriptor and
        second element is for angular descriptor. Default is [8, 8].
    l_max: list[int, int, int]
       The maximum expansion order for the angular terms. First element is for
       three-body, second element is for four-body and third element is for five-body.
       Default is [4, 2, 1].
    neuron: int
        Number of neurons in the hidden layer. Default is 80.
    lambda_1: float
        Weight for L1 regularization. Default is 0.
    lambda_e: float
        Weight for energy loss. Default is 1.
    lambda_f: float
        Weight for force loss. Default is 1.
    lambda_v: float
        Weight for virial loss. Default is 0.1.
    force_delta: float
        Sets bias the on the loss function to put more emphasis on obtaining
        accurate predictions for smaller forces. Default is 0.
    batch: int
        Batch size for training. Default is 1000.
    population: int
        Size of the population used by the SNES algorithm. Default is 50.
    generation: bool
        Sets the max number of generations for SNES algorithm.
    zbl : float
        Cutoff to use in universal ZBL potential at short distances.
        Acceptable values are in range 1 to 2.5. Default is 2.

    References
    ----------
    * GPUMD & NEP: https://doi.org/10.1063/5.0106617.
    * SNES : https://doi.org/10.1145/2001576.2001692.
    * Parameter defaults taken from SI: https://doi.org/10.1038/s41467-024-54554-x.

    Returns
    -------
    dict[str, float]
        A dictionary mapping 'train_error', 'test_error', and 'mlip_path'.
    """
    hyperparameters = hyperparameters.model_copy(deep=True)

    train_data = ase.io.read(os.path.join(db_dir, "train.extxyz"), index=":")
    test_data = ase.io.read(os.path.join(db_dir, "test.extxyz"), index=":")

    try:
        train_nep = [
            at for at in train_data if "IsolatedAtom" not in at.info["config_type"]
        ]
        test_nep = [
            at for at in test_data if "IsolatedAtom" not in at.info["config_type"]
        ]
    except KeyError:
        train_nep = train_data
        test_nep = test_data

    # Use the SinglePointCalculator to set the energy, forces, and virial
    # Step required to generate NEP compatible xyz file using write_structures from calorine
    for at in train_data:
        at.calc = SinglePointCalculator(
            at, energy=at.info[ref_energy_name], forces=at.arrays[ref_force_name]
        )
        at.info["virial"] = at.info[ref_virial_name]
        del at.info[ref_energy_name]
        del at.info[ref_virial_name]
        del at.arrays[ref_force_name]

    for at in test_data:
        at.calc = SinglePointCalculator(
            at, energy=at.info[ref_energy_name], forces=at.arrays[ref_force_name]
        )
        at.info["virial"] = at.info[ref_virial_name]
        del at.info[ref_energy_name]
        del at.info[ref_virial_name]
        del at.arrays[ref_force_name]

    write_structures(outfile="train.xyz", structures=train_nep)
    write_structures(outfile="test.xyz", structures=test_nep)

    if fit_kwargs:
        hyperparameters.update_parameters(fit_kwargs)

    nep_hypers = hyperparameters.model_dump(by_alias=True)

    nep_hypers["type"] = [len(species_list), *species_list]
    nep_hypers["type_weight"] = [1.0] * len(species_list)

    write_nepfile(parameters=nep_hypers, dirname=".")
    run_nep(gpu_identifier_indices=gpu_identifier_indices)

    metrics_df = read_loss("loss.out")

    return {
        "train_error": metrics_df.RMSE_E_train.values[-1],
        "test_error": metrics_df.RMSE_E_test.values[-1],
        "mlip_path": Path.cwd(),
    }


def nequip_fitting(
    db_dir: Path,
    hyperparameters: NEQUIP_HYPERS = NEQUIP_HYPERS,
    isolated_atom_energies: dict | None = None,
    ref_energy_name: str = "REF_energy",
    ref_force_name: str = "REF_forces",
    ref_virial_name: str = "REF_virial",
    fit_kwargs: dict | None = None,
    device: str = "cuda",
) -> dict:
    """
    Perform the NequIP potential fitting.

    This function sets up and executes a python script to perform NequIP fitting using specified parameters
    and input data located in the provided directory. It handles the input/output of atomic configurations,
    sets up the NequIP model, and calculates training and testing errors after fitting.

    Parameters
    ----------
    db_dir: Path
        directory containing the training and testing data files.
    hyperparameters: MLIP_HYPERS.NEQUIP
        Fit hyperparameters.
    isolated_atom_energies: dict
        mandatory dictionary mapping element numbers to isolated energies.
    ref_energy_name : str, optional
        Reference energy name.
    ref_force_name : str, optional
        Reference force name.
    ref_virial_name : str, optional
        Reference virial name.
    device: str
        specify device to use cuda or cpu
    fit_kwargs: dict.
        optional dictionary with parameters for nequip fitting with keys same as
        mlip-rss-defaults.json.

    Keyword Arguments
    -----------------
    r_max: float
        cutoff radius in length units
    num_layers: int
        number of interaction blocks
    l_max: int
        maximum irrep order (rotation order) for the network's features
    num_features: int
        multiplicity of the features
    num_basis: int
        number of basis functions used in the radial basis
    invariant_layers: int
        number of radial layers
    invariant_neurons: int
        number of hidden neurons in radial function
    batch_size: int
        batch size
    learning_rate: float
        learning rate
    default_dtype: str
        type of float to use, e.g. float32 and float64

    Returns
    -------
    dict[str, float]
        A dictionary containing train_error, test_error, and the path to the fitted MLIP.

    Raises
    ------
    - ValueError: If the `isolated_atom_energies` dictionary is empty or not provided when required.
    """
    """
    [TODO] train Nequip on virials
    """
    hyperparameters = hyperparameters.model_copy(deep=True)

    train_data = ase.io.read(os.path.join(db_dir, "train.extxyz"), index=":")
    train_nequip = [
        at for at in train_data if "IsolatedAtom" not in at.info["config_type"]
    ]
    ase.io.write("train_nequip.extxyz", train_nequip, format="extxyz")

    test_data = ase.io.read(os.path.join(db_dir, "test.extxyz"), index=":")
    num_of_train = len(train_nequip)
    num_of_val = len(test_data)

    isolated_atom_energies_update = ""
    ele_syms = []
    if isolated_atom_energies:
        for e_num in isolated_atom_energies:
            element_symbol = "  - " + chemical_symbols[int(e_num)] + "\n"
            isolated_atom_energies_update += element_symbol
            ele_syms.append(chemical_symbols[int(e_num)])
    else:
        raise ValueError("isolated_atom_energies is empty or not defined!")

    nequip_config_updates = {
        "dataset_key_mapping": {
            f"{ref_energy_name}": "total_energy",
            f"{ref_force_name}": "forces",
        },
        "validation_dataset_key_mapping": {
            f"{ref_energy_name}": "total_energy",
            f"{ref_force_name}": "forces",
        },
        "chemical_symbols": ele_syms,
        "dataset_file_name": "./train_nequip.extxyz",
        "validation_dataset_file_name": f"{db_dir}/test.extxyz",
        "n_train": num_of_train,
        "n_val": num_of_val,
    }
    hyperparameters.update_parameters(nequip_config_updates)

    if fit_kwargs:
        hyperparameters.update_parameters(fit_kwargs)

    nequip_hypers = hyperparameters.model_dump(by_alias=True)

    dumpfn(nequip_hypers, "nequip.yaml")

    run_nequip("nequip-train nequip.yaml", "nequip_train")
    run_nequip(
        "nequip-deploy build --train-dir results/autoplex ./deployed_nequip_model.pth",
        "nequip_deploy",
    )

    calc = NequIPCalculator.from_deployed_model(
        model_path="deployed_nequip_model.pth",
        device=device,
        species_to_type_name={s: s for s in ele_syms},
        set_global_options=False,
    )

    ener_out_train = []
    for at in train_nequip:
        at.calc = calc
        ener_out_train.append(at.get_potential_energy() / len(at))

    ener_in_train = [
        at.info["REF_energy"] / len(at.get_chemical_symbols()) for at in train_nequip
    ]

    train_error = rms_dict(ener_in_train, ener_out_train)["rmse"]

    ener_out_test = []
    for at in test_data:
        at.calc = calc
        ener_out_test.append(at.get_potential_energy() / len(at))

    ener_in_test = [
        at.info["REF_energy"] / len(at.get_chemical_symbols()) for at in test_data
    ]

    test_error = rms_dict(ener_in_test, ener_out_test)["rmse"]

    return {
        "train_error": train_error,
        "test_error": test_error,
        "mlip_path": Path.cwd(),
    }


def m3gnet_fitting(
    db_dir: Path,
    hyperparameters: M3GNET_HYPERS = M3GNET_HYPERS,
    device: str = "cuda",
    ref_energy_name: str = "REF_energy",
    ref_force_name: str = "REF_forces",
    ref_virial_name: str = "REF_virial",
    test_equal_to_val: bool = True,
    fit_kwargs: dict | None = None,
) -> dict:
    """
    Perform the M3GNet potential fitting.

    Parameters
    ----------
    db_dir: Path
        Directory containing the training and testing data files.
    hyperparameters: MLIP_HYPERS.M3GNET
        Fit hyperparameters.
    device: str
        Device on which the model will be trained, e.g., 'cuda' or 'cpu'.
    ref_energy_name : str, optional
        Reference energy name.
    ref_force_name : str, optional
        Reference force name.
    ref_virial_name : str, optional
        Reference virial name.
    test_equal_to_val: bool
        If True, the testing dataset will be the same as the validation dataset.
    fit_kwargs: dict.
        optional dictionary with parameters for M3GNET fitting.

    Keyword Arguments
    -----------------
    exp_name: str
        Name of the experiment, used for saving model checkpoints and logs.
    results_dir: str
        Directory to store the training results and fitted model.
    cutoff: float
        Cutoff radius for atomic interactions in length units.
    threebody_cutoff: float
        Cutoff radius for three-body interactions in length units.
    batch_size: int
        Number of structures per batch during training.
    max_epochs: int
        Maximum number of training epochs.
    include_stresses: bool
        If True, includes stress tensors in the model predictions and training process.
    dim_node_embedding: int
         Dimension of node embedding.
    dim_edge_embedding: int
        Dimension of edge embeddings.
    units: int
        Number of units in each dense layer of the model.
    max_l: int
        Maximum degree of spherical harmonics.
    max_n: int
        Maximum radial function degree.

    Returns
    -------
    dict[str, float]
        A dictionary containing keys such as 'train_error', 'test_error', and 'path_to_fitted_model',
        representing the training error, test error, and the location of the saved model, respectively.

    References
    ----------
    *    Title: Tutorials of Materials Graph Library (MatGL)
    *    Author: Tsz Wai Ko, Chi Chen and Shyue Ping Ong
    *    Version: 1.1.3
    *    Date 7/8/2024
    *    Availability: https://matgl.ai/tutorials%2FTraining%20a%20M3GNet%20Potential%20with%20PyTorch%20Lightning.html
    *    License: BSD 3-Clause License
    """
    hyperparameters = hyperparameters.model_copy(deep=True)

    if fit_kwargs:
        hyperparameters.update_parameters(fit_kwargs)

    m3gnet_hypers = hyperparameters.model_dump(by_alias=True)

    exp_name = m3gnet_hypers["exp_name"]
    results_dir = m3gnet_hypers["results_dir"]

    os.makedirs(os.path.join(results_dir, exp_name), exist_ok=True)

    with open("output.txt", "w") as f:
        # Backup original stdout stream.
        original_stdout = sys.stdout

        # Set stdout to the file object.
        sys.stdout = f

        # Print something (it goes to the file).
        print("This line will be written to the file.")

        # Restore original stdout stream.
        sys.stdout = original_stdout

    with open("m3gnet.log", "w") as log_file:
        original_stdout = sys.stdout
        original_stderr = sys.stderr
        sys.stdout = log_file
        sys.stderr = log_file

        train_data = ase.io.read(os.path.join(db_dir, "train.extxyz"), index=":")
        train_m3gnet = [
            at
            for at in train_data
            if "IsolatedAtom" not in at.info["config_type"]
            and "dimer" not in at.info["config_type"]
        ]

        # prepare train dataset
        (
            train_structs,
            train_energies,
            train_forces,
            train_stresses,
        ) = convert_xyz_to_structure(
            train_m3gnet,
            include_forces=True,
            include_stresses=m3gnet_hypers.get("include_stresses"),
            ref_energy_name=ref_energy_name,
            ref_force_name=ref_force_name,
            ref_virial_name=ref_virial_name,
        )

        train_labels = {
            "energies": train_energies,
            "forces": train_forces,
            "stresses": train_stresses,
        }
        train_element_types = get_element_list(train_structs)

        print(
            train_element_types
        )  # this print has to stay as the stdout is written to the file
        train_converter = Structure2Graph(
            element_types=train_element_types, cutoff=m3gnet_hypers.get("cutoff")
        )
        train_datasets = MGLDataset(
            threebody_cutoff=m3gnet_hypers.get("threebody_cutoff"),
            structures=train_structs,
            converter=train_converter,
            labels=train_labels,
            include_line_graph=True,
            filename="dgl_graph_train.bin",
            filename_lattice="lattice_train.pt",
            filename_line_graph="dgl_line_graph_train.bin",
            filename_state_attr="state_attr_train.pt",
            filename_labels="labels_train.json",
            save_dir=os.path.join(results_dir, exp_name),
        )

        if os.path.exists(os.path.join(db_dir, "test.extxyz")):
            test_data = ase.io.read(os.path.join(db_dir, "test.extxyz"), index=":")
            # prepare test dataset
            (
                test_structs,
                test_energies,
                test_forces,
                test_stresses,
            ) = convert_xyz_to_structure(
                test_data,
                include_forces=True,
                include_stresses=m3gnet_hypers.get("include_stresses"),
                ref_energy_name=ref_energy_name,
                ref_force_name=ref_force_name,
                ref_virial_name=ref_virial_name,
            )

            test_labels = {
                "energies": test_energies,
                "forces": test_forces,
                "stresses": test_stresses,
            }
            test_element_types = get_element_list(test_structs)
            test_converter = Structure2Graph(
                element_types=test_element_types, cutoff=m3gnet_hypers.get("cutoff")
            )
            test_dataset = MGLDataset(
                threebody_cutoff=m3gnet_hypers.get("threebody_cutoff"),
                structures=test_structs,
                converter=test_converter,
                labels=test_labels,
                include_line_graph=True,
                filename="dgl_graph_test.bin",
                filename_lattice="lattice_test.pt",
                filename_line_graph="dgl_line_graph_test.bin",
                filename_state_attr="state_attr_test.pt",
                filename_labels="labels_test.json",
                save_dir=os.path.join(results_dir, exp_name),
            )

        if test_equal_to_val:
            train_dataset = train_datasets
            val_dataset = test_dataset
        else:
            if os.path.exists(os.path.join(db_dir, "test.extxyz")):
                train_dataset, val_dataset, _ = split_dataset(
                    train_datasets,
                    frac_list=[0.9, 0.1, 0],  # to guarantee train:valid=9:1
                    shuffle=True,
                    random_state=42,
                )
            else:
                train_dataset, val_dataset, test_dataset = split_dataset(
                    train_datasets,
                    frac_list=[0.8, 0.1, 0.1],  # to guarantee train:valid:test=8:1:1
                    shuffle=True,
                    random_state=42,
                )

        my_collate_fn = partial(
            collate_fn_pes, include_line_graph=True
        )  # Set all include_line_graph to False will disable three-body interactions
        train_loader, val_loader, test_loader = MGLDataLoader(
            train_data=train_dataset,
            val_data=val_dataset,
            test_data=test_dataset,
            collate_fn=my_collate_fn,
            batch_size=m3gnet_hypers.get("batch_size"),
            num_workers=1,
        )
        # train from scratch
        if not m3gnet_hypers["foundation_model"]:  # train from scratch
            model = M3GNet(
                element_types=train_element_types,
                is_intensive=m3gnet_hypers.get("is_intensive"),
                cutoff=m3gnet_hypers.get("cutoff"),
                threebody_cutoff=m3gnet_hypers.get("threebody_cutoff"),
                dim_node_embedding=m3gnet_hypers.get("dim_node_embedding"),
                dim_edge_embedding=m3gnet_hypers.get("dim_edge_embedding"),
                units=m3gnet_hypers.get("units"),
                max_l=m3gnet_hypers.get("max_l"),
                max_n=m3gnet_hypers.get("max_n"),
                nblocks=m3gnet_hypers.get("nblocks"),
            )
            lit_module = PotentialLightningModule(
                model=model,
                element_refs=m3gnet_hypers.get("element_refs"),
                include_line_graph=m3gnet_hypers.get("include_line_graph"),
                allow_missing_labels=m3gnet_hypers.get("allow_missing_labels"),
                energy_weight=m3gnet_hypers.get("energy_weight"),
                force_weight=m3gnet_hypers.get("force_weight"),
                lr=m3gnet_hypers.get("lr"),
                loss=m3gnet_hypers.get("loss"),
                loss_params=m3gnet_hypers.get("loss_params"),
                stress_weight=m3gnet_hypers.get("stress_weight"),
                magmom_weight=m3gnet_hypers.get("magmom_weight"),
                data_mean=m3gnet_hypers.get("data_mean"),
                data_std=m3gnet_hypers.get("data_std"),
                decay_alpha=m3gnet_hypers.get("decay_alpha"),
                decay_steps=m3gnet_hypers.get("decay_steps"),
                sync_dist=m3gnet_hypers.get("sync_dist"),
                magmom_target=m3gnet_hypers.get("magmom_target"),
                optimizer=m3gnet_hypers.get("optimizer"),
                scheduler=m3gnet_hypers.get("scheduler"),
            )
        else:  # finetune a foundation model (pretrained model)
            logging.info(
                f"Finetuning foundation model: {m3gnet_hypers['foundation_model']}"
            )
            m3gnet_nnp = matgl.load_model(m3gnet_hypers["foundation_model"])
            model = m3gnet_nnp.model
            property_offset = (
                m3gnet_nnp.element_refs.property_offset
                if m3gnet_hypers["use_foundation_model_element_refs"]
                else None
            )
            lit_module = PotentialLightningModule(
                model=model,
                element_refs=property_offset,
                include_line_graph=m3gnet_hypers.get("include_line_graph"),
                allow_missing_labels=m3gnet_hypers.get("allow_missing_labels"),
                energy_weight=m3gnet_hypers.get("energy_weight"),
                force_weight=m3gnet_hypers.get("force_weight"),
                lr=m3gnet_hypers.get("lr"),
                loss=m3gnet_hypers.get("loss"),
                loss_params=m3gnet_hypers.get("loss_params"),
                stress_weight=m3gnet_hypers.get("stress_weight"),
                magmom_weight=m3gnet_hypers.get("magmom_weight"),
                data_mean=m3gnet_hypers.get("data_mean"),
                data_std=m3gnet_hypers.get("data_std"),
                decay_alpha=m3gnet_hypers.get("decay_alpha"),
                decay_steps=m3gnet_hypers.get("decay_steps"),
                sync_dist=m3gnet_hypers.get("sync_dist"),
                magmom_target=m3gnet_hypers.get("magmom_target"),
                optimizer=m3gnet_hypers.get("optimizer"),
                scheduler=m3gnet_hypers.get("scheduler"),
            )

        logger = CSVLogger(name=exp_name, save_dir=os.path.join(results_dir, "logs"))
        # Inference mode = False is required for calculating forces, stress in test mode and prediction mode
        if device == "cuda":
            if torch.cuda.is_available():
                gpu_id = os.environ.get("CUDA_VISIBLE_DEVICES", "0")
                torch.cuda.set_device(torch.device(f"cuda:{gpu_id}"))
                trainer = pl.Trainer(
                    max_epochs=m3gnet_hypers.get("max_epochs"),
                    accelerator="gpu",
                    logger=logger,
                    inference_mode=False,
                )
            else:
                raise ValueError("CUDA is not available.")
        else:
            trainer = pl.Trainer(
                max_epochs=m3gnet_hypers.get("max_epochs"),
                accelerator="cpu",
                logger=logger,
                inference_mode=False,
            )
        # Again loggers ...
        print("Start training...")
        print(f"Length of train_loader: {len(train_loader)}")
        print(f"Length of val_loader: {len(val_loader)}")
        print(f"Length of test_loader: {len(test_loader)}")
        trainer.fit(
            model=lit_module, train_dataloaders=train_loader, val_dataloaders=val_loader
        )
        # test the model, remember to set inference_mode=False in trainer (see above)
        print("Train error:")
        trainer.test(dataloaders=train_loader)
        print("Valid error:")
        trainer.test(dataloaders=val_loader)
        print("Test error:")
        trainer.test(dataloaders=test_loader)

        # save trained model
        model_export_path = os.path.join(results_dir, exp_name)
        # model.save(model_export_path)
        potential = Potential(model=model)
        potential.save(model_export_path)

        sys.stdout = original_stdout
        sys.stderr = original_stderr

    for fn in (
        "dgl_graph_train.bin",
        "lattice_train.pt",
        "dgl_line_graph_train.bin",
        "state_attr_train.pt",
        "labels_train.json",
        "dgl_graph_test.bin",
        "lattice_test.pt",
        "dgl_line_graph_test.bin",
        "state_attr_test.pt",
        "labels_test.json",
    ):
        with contextlib.suppress(FileNotFoundError):
            os.remove(os.path.join(results_dir, exp_name, fn))

    sections = {
        "Train error:": {
            "train_Energy_RMSE": "test_Energy_RMSE",
            "train_Force_RMSE": "test_Force_RMSE",
        },
        "Valid error:": {
            "val_Energy_RMSE": "test_Energy_RMSE",
            "val_Force_RMSE": "test_Force_RMSE",
        },
        "Test error:": {
            "test_Energy_RMSE": "test_Energy_RMSE",
            "test_Force_RMSE": "test_Force_RMSE",
        },
    }

    extracted_values = {}
    with open("m3gnet.log") as file:
        content = file.read()

        for section, metrics in sections.items():
            start_index = content.find(section)
            if start_index != -1:
                next_index = min(
                    [
                        content.find(sec, start_index + 1)
                        for sec in sections
                        if content.find(sec, start_index + 1) != -1
                    ],
                    default=len(content),
                )
                section_content = content[start_index:next_index]
                for key, metric in metrics.items():
                    for line in section_content.split("\n"):
                        if metric in line:
                            if metric in line.split()[0]:
                                extracted_values[key] = float(line.split()[1])
                            else:
                                extracted_values[key] = float(line.split()[3])

    for key, value in extracted_values.items():
        print(f"{key}: {value}")

    """
    !!![Note] The RMSE directly outputted from Torch is not strictly the RMSE of the full datasets;
    it is related to the batch size. It only becomes a strict RMSE when the batch size is larger
    than the size of the dataset. The output here can be considered as an approximate result.
    [TODO] Switch it to the strict RMSE.
    """
    mlip_path = Path.cwd() / model_export_path

    return {
        "train_error": extracted_values["train_Energy_RMSE"],
        "test_error": extracted_values["test_Energy_RMSE"],
        "mlip_path": mlip_path,
    }


def mace_fitting(
    db_dir: Path,
    hyperparameters: MACE_HYPERS = MACE_HYPERS,
    device: Literal["cpu", "cuda", "mps", "xpu"] = Field(
        default="cpu", description="Device to be used for model fitting"
    ),
    ref_energy_name: str = "REF_energy",
    ref_force_name: str = "REF_forces",
    ref_virial_name: str = "REF_virial",
    fit_kwargs: dict | None = None,
) -> dict:
    """
    Perform the MACE potential fitting.

    This function sets up and executes a python script to perform MACE fitting using specified parameters
    and input data located in the provided directory. It handles the input/output of atomic configurations,
    sets up the NequIP model, and calculates training and testing errors after fitting.

    Parameters
    ----------
    db_dir: Path
        directory containing the training and testing data files.
    hyperparameters: MLIP_HYPERS.MACE
        Fit hyperparameters.
    device: str
        specify device to use cuda or cpu.
    ref_energy_name : str, optional
        Reference energy name.
    ref_force_name : str, optional
        Reference force name.
    ref_virial_name : str, optional
        Reference virial name.
    fit_kwargs: dict.
        optional dictionary with parameters for mace fitting with keys same as
        mlip-rss-defaults.json.

    Keyword Arguments
    -----------------
    model: str
        type of model to be trained
    config_type_weights: str
        weights of config types
    hidden_irreps: str
        control the model size
    r_max: float
        cutoff radius controls the locality of the model
    batch_size: int
        batch size (note that batch size cannot be larger than the size of training datasets)
    start_swa: str
        if the keyword --swa is enabled, the energy weight of the loss is increased
        for the last ~20% of the training epochs (from --start_swa epochs)
    correlation: int
        correlation order corresponds to the order that MACE induces at each layer
    loss: str
        loss functions
    default_dtype: str
        type of float to use, e.g. float32 and float64

    Returns
    -------
    dict[str, float]
        A dictionary containing train_error, test_error, and the path to the fitted MLIP.

    """
    hyperparameters = hyperparameters.model_copy(deep=True)

    if ref_virial_name is not None:
        atoms = read(f"{db_dir}/train.extxyz", index=":")
        mace_convert_virial_to_stress(
            atoms=atoms, ref_virial_name=ref_virial_name, out_file_name="train.extxyz"
        )

    hyperparameters.update_parameters(fit_kwargs)

    mace_hypers = hyperparameters.model_dump(by_alias=True, exclude_none=True)

    boolean_hypers = [
        "distributed",
        "pair_repulsion",
        "amsgrad",
        "swa",
        "stage_two",
        "keep_checkpoint",
        "save_all_checkpoints",
        "restart_latest",
        "save_cpu",
        "wandb",
        "compute_statistics",
        "foundation_model_readout",
        "ema",
    ]
    boolean_str_hypers = [
        "compute_avg_num_neighbors",
        "compute_stress",
        "compute_forces",
        "multi_processed_test",
        "pin_memory",
        "foundation_filter_elements",
        "multiheads_finetuning",
        "keep_isolated_atoms",
        "shuffle",
    ]

    hypers = []
    for hyper in mace_hypers:
        if hyper in boolean_hypers:
            if mace_hypers[hyper] is True:
                hypers.append(f"--{hyper}")
        elif hyper in boolean_str_hypers:
            hypers.append(f"--{hyper}={mace_hypers[hyper]}")
        elif hyper in ["train_file", "test_file"]:
            logging.info("Train and test files have default names.")
        elif hyper in ["energy_key", "virial_key", "forces_key", "device"]:
            logging.info("energy_key, virial_key and forces_key have default names.")
        else:
            hypers.append(f"--{hyper}={mace_hypers[hyper]}")

    hypers.append(f"--train_file={db_dir}/train.extxyz")
    hypers.append(f"--valid_file={db_dir}/test.extxyz")

    if ref_energy_name is not None:
        hypers.append(f"--energy_key={ref_energy_name}")
    if ref_force_name is not None:
        hypers.append(f"--forces_key={ref_force_name}")
    if ref_virial_name is not None:
        hypers.append(
            "--stress_key={'REF_stress'}"
        )  # MACE will be trained on stress instead of virial.
        # They are essentially equivalent, but since the MACE-torch log file directly
        # reports the stress error, we train on stress for consistency.
    if device is not None:
        hypers.append(f"--device={device}")

    run_mace(hypers)

    try:
        with open("./logs/MACE_model_run-123.log") as file:
            log_data = file.read()
    except FileNotFoundError:
        # to cover finetuning
        try:
            with open(f"./logs/{fit_kwargs['name']}_run-123.log") as file:
                log_data = file.read()
        except FileNotFoundError:
            try:
                with open("./logs/MACE_final_run-3.log") as file:
                    log_data = file.read()
            except FileNotFoundError:
                with open(f"./logs/{fit_kwargs['name']}_run-3.log") as file:
                    log_data = file.read()

    tables = re.split(r"\+-+\+\n", log_data)
    # if tables:
    last_table = tables[-2]
    try:
        matches = re.findall(
            r"\|\s*(train_default|valid_default)\s*\|\s*([\d\.]+)\s*\|",
            last_table,
            re.IGNORECASE,
        )

        return {
            "train_error": float(matches[0][1]),
            "test_error": float(matches[1][1]),
            "mlip_path": Path.cwd(),
        }
    except IndexError:
        # to ensure backward compatibility to mace 0.3.4
        matches = re.findall(r"\|\s*(train|valid)\s*\|\s*([\d\.]+)\s*\|", last_table)

        return {
            "train_error": float(matches[0][1]),
            "test_error": float(matches[1][1]),
            "mlip_path": Path.cwd(),
        }


def check_convergence(test_error: float) -> bool:
    """
    Check the convergence of the fit.

    Parameters
    ----------
    test_error:
        The error of the test data.

    Returns
    -------
    The convergence bool.
    """
    convergence = False
    if test_error < 0.01:
        convergence = True

    return convergence


def gap_hyperparameter_constructor(
    gap_parameter_dict: dict,
    include_two_body: bool = False,
    include_three_body: bool = False,
    include_soap: bool = False,
) -> list:
    """
    Construct a list of arguments needed to execute gap potential from the parameters' dict.

    Parameters
    ----------
    gap_parameter_dict : dict.
        dictionary with gap hyperparameters.
    include_two_body : bool.
        bool indicating whether to include two-body hyperparameters
    include_three_body : bool.
        bool indicating whether to include three-body hyperparameters
    include_soap : bool.
        bool indicating whether to include soap hyperparameters

    Returns
    -------
        list
           gap fit input parameter string.
    """
    dict_wo_term_name = gap_parameter_dict.copy()
    if "two_body" in dict_wo_term_name["general"]:
        del dict_wo_term_name["general"]["two_body"]
    if "three_body" in dict_wo_term_name["general"]:
        del dict_wo_term_name["general"]["three_body"]
    if "soap" in dict_wo_term_name["general"]:
        del dict_wo_term_name["general"]["soap"]

    general = [f"{key}={value}" for key, value in dict_wo_term_name["general"].items()]

    two_body_params = " ".join(
        [
            f"{key}={value}"
            for key, value in dict_wo_term_name["twob"].items()
            if include_two_body is True
        ]
    )

    three_body_params = " ".join(
        [
            f"{key}={value}"
            for key, value in dict_wo_term_name["threeb"].items()
            if include_three_body is True
        ]
    )

    soap_params = " ".join(
        [
            f"{key}={value}"
            for key, value in dict_wo_term_name["soap"].items()
            if include_soap is True
        ]
    )
    # add separator between the arg types
    if include_two_body and include_three_body and include_soap:
        three_body_params = " :" + three_body_params
        soap_params = " :soap " + soap_params
    elif include_two_body and include_three_body and not include_soap:
        three_body_params = " :" + three_body_params
    elif (include_two_body or include_three_body) and include_soap:
        soap_params = " :soap " + soap_params
    elif include_soap and not include_three_body and not include_two_body:
        soap_params = "soap " + soap_params

    gap_hyperparameters = f"gap={{{two_body_params}{three_body_params}{soap_params}}}"

    return [*general, gap_hyperparameters]


def get_list_of_vasp_calc_dirs(flow_output) -> list[str]:
    """
    Return a list of vasp_calc_dirs from PhononDFTMLDataGenerationFlow output.

    Parameters
    ----------
    flow_output: dict.
        PhononDFTMLDataGenerationFlow output

    Returns
    -------
    list.
        A list of vasp_calc_dirs
    """
    list_of_vasp_calc_dirs: list[str] = []
    for output in flow_output.values():
        for output_type, dirs in output.items():
            if output_type != "phonon_data" and isinstance(dirs, list):
                if output_type == "rattled_dir":
                    flat_dirs = [[item for sublist in dirs for item in sublist]]
                    list_of_vasp_calc_dirs.extend(*flat_dirs)
                else:
                    list_of_vasp_calc_dirs.extend(*dirs)

    return list_of_vasp_calc_dirs


def vaspoutput_2_extended_xyz(
    path_to_vasp_static_calcs: list,
    config_types: list[str] | None = None,
    data_types: list[str] | None = None,
    ref_energy_name: str = "REF_energy",
    ref_force_name: str = "REF_forces",
    ref_virial_name: str = "REF_virial",
    regularization: float = 0.1,
    f_min: float = 0.01,  # unit: eV Å-1
    atom_wise_regularization: bool = True,
) -> None:
    """
    Parse all VASP output files (vasprun.xml/OUTCAR) and generates a vasp_ref.extxyz.

    Uses ase.io.read to parse the OUTCARs
    Adapted from https://lipai.github.io/scripts/ml_scripts/outcar2xyz.html

    Parameters
    ----------
    path_to_vasp_static_calcs : list.
        List of VASP static calculation directories.
    config_types: list[str] or None
            list of config_types.
    data_types: list[str] or None
            track the data type (phonon or random).
    ref_energy_name : str
        Reference energy name in xyz file.
    ref_force_name : str
        Reference force name in xyz file.
    ref_virial_name : str
        Reference virial name in xyz file.
    regularization: float
        regularization value for the atom-wise force components.
    f_min: float
        minimal force cutoff value for atom-wise regularization.
    atom_wise_regularization: bool
        for including atom-wise regularization.
    """
    counter = 0
    if config_types is None:
        config_types = ["bulk"] * len(path_to_vasp_static_calcs)
    if data_types is None:
        data_types = ["other"] * len(path_to_vasp_static_calcs)

    for path, config_type, data_type in zip(
        path_to_vasp_static_calcs, config_types, data_types
    ):
        # strip hostname if it exists in the path
        path_without_hostname = Path(strip_hostname(path)).joinpath("vasprun.xml.gz")
        path_without_hostname2 = Path(strip_hostname(path)).joinpath(
            "final_atoms_object.xyz"
        )
        path_without_hostname2 = Path(strip_hostname(path)).joinpath(
            "final_atoms_object.xyz"
        )
        try:
            # read the vasp output
            if path_without_hostname.exists():
                file = read(path_without_hostname, index=":")
            elif path_without_hostname2.exists() or path_without_hostname2.exists():
                file = read(path_without_hostname2, index=":")
            for i in file:
                virial_list = (
                    -voigt_6_to_full_3x3_stress(i.get_stress()) * i.get_volume()
                )
                i.info[ref_virial_name] = " ".join(map(str, virial_list.flatten()))
                del i.calc.results["stress"]
                i.arrays[ref_force_name] = i.calc.results["forces"]
                if atom_wise_regularization and (data_type == "phonon_dir"):
                    atom_forces = np.array(i.arrays[ref_force_name])
                    atom_wise_force = np.array(
                        [
                            force if force > f_min else f_min
                            for force in np.linalg.norm(atom_forces, axis=1)
                        ]
                    )
                    i.arrays["force_atom_sigma"] = regularization * atom_wise_force
                del i.calc.results["forces"]
                i.info[ref_energy_name] = i.calc.results["free_energy"]
                del i.calc.results["energy"]
                del i.calc.results["free_energy"]
                i.info["config_type"] = config_type
                i.info["data_type"] = data_type.rstrip("_dir")
                i.pbc = True

            # TODO: maybe only add isolated atoms energy if it wasn't there?

            write("vasp_ref.extxyz", file, append=True)

        except FileNotFoundError:
            counter += 1

        if counter / len(path_to_vasp_static_calcs) > 0.05:
            raise ValueError(
                "An insufficient number of data points collected. Workflow stopped."
            )


def flatten(atoms_object, recursive=False) -> list[str | bytes | Atoms] | list:
    """
    Flatten an iterable fully, but excluding Atoms objects.

    Parameters
    ----------
    atoms_object: Atoms object
    recursive: bool
        set the recursive boolean.

    Returns
    -------
    a flattened object, excluding the Atoms objects.

    """
    iteration_list: list[str | bytes | Atoms] | list = []

    if recursive:
        for element in atoms_object:
            if isinstance(element, Iterable) and not isinstance(
                element, (str, bytes, ase.atoms.Atoms, ase.Atoms)
            ):
                iteration_list.extend(flatten(element, recursive=True))
            else:
                iteration_list.append(element)
        return iteration_list

    return [item for sublist in atoms_object for item in sublist]


def gcm3_to_Vm(gcm3, mr, n_atoms=1) -> float:
    """
    Convert gcm3 to Vm.

    Parameters
    ----------
    gcm3:
        Density in grams per cubic centimeter (g/cm³).
    mr:
        Molar mass in grams per mole (g/mol).
    n_atoms:
        Number of atoms in the formula unit. Default is 1.

    Returns
    -------
    the converted unit.

    """
    return 1 / (n_atoms * (gcm3 / mr) * 6.022e23 / (1e8) ** 3)


def get_atomic_numbers(species: list) -> list[int]:
    """
    Get atomic numbers.

    Parameters
    ----------
    species:
        type of species

    Returns
    -------
    atomic_numbers:
        list of atomic numbers.

    """
    atom_numbers = []
    for atom_type in species:
        atom = Atoms(atom_type, [(0, 0, 0)])
        atom_numbers.append(int(atom.get_atomic_numbers()[0]))

    return atom_numbers


def energy_remain(in_file: str) -> float:
    """
    Plot the distribution of energy per atom on the output vs. the input.

    Parameters
    ----------
    in_file:
        input file

    Returns
    -------
    rms["rmse"]:
        distribution of energy per atom RMSE of output vs. input.

    """
    # read files
    in_atoms = ase.io.read(in_file, ":")
    if "config_type" in in_atoms[0].info:
        ener_in = [
            at.info["REF_energy"] / len(at.get_chemical_symbols())
            for at in in_atoms
            if at.info["config_type"] != "IsolatedAtoms"
        ]
        ener_out = [
            at.get_potential_energy() / len(at.get_chemical_symbols())
            for at in in_atoms
            if at.info["config_type"] != "IsolatedAtoms"
        ]
    else:
        ener_in = [
            at.info["REF_energy"] / len(at.get_chemical_symbols()) for at in in_atoms
        ]
        ener_out = [
            at.get_potential_energy() / len(at.get_chemical_symbols())
            for at in in_atoms
        ]
    rms = rms_dict(ener_in, ener_out)
    return rms["rmse"]


def extract_gap_label(xml_file_path) -> str:
    """
    Extract GAP label.

    Parameters
    ----------
    xml_file_path:
        path to the GAP fit potential xml file.

    Returns
    -------
    the extracted GAP label.

    """
    # Parse the XML file
    tree = ET.parse(xml_file_path)
    root = tree.getroot()
    return root.tag


def plot_convex_hull(all_points: np.ndarray, hull_points: np.ndarray) -> None:
    """
    Plot convex hull.

    Parameters
    ----------
    all_points : np.ndarray
        Array of all points to be plotted.
    hull_points : np.ndarray
        Array of points used to calculate the convex hull.
    """
    hull = ConvexHull(hull_points)

    plt.plot(all_points[:, 0], all_points[:, 1], "o", markersize=3, label="All Points")

    for i, simplex in enumerate(hull.simplices):
        if i == 0:
            plt.plot(
                hull_points[simplex, 0],
                hull_points[simplex, 1],
                "k-",
                label="Convex Hull",
            )
        else:
            plt.plot(hull_points[simplex, 0], hull_points[simplex, 1], "k-")

    plt.xlabel("Volume")
    plt.ylabel("Energy")
    plt.title("Convex Hull with All Points")
    plt.legend()
    plt.savefig("ConvexHull.png")


def calculate_delta_2b(
    atoms_db: list[Atoms], cutoff: float, e_name: str
) -> tuple[float, ndarray]:
    """
    Calculate the delta parameter and average number of triplets for gap-fitting.

    Parameters
    ----------
    atoms_db: list[Atoms]
        list of Ase atoms objects
    nb: int
        Two-body or three-body interactions.
    cutoff: float
        Cutoff radius used to compute the dimensionality of the descriptor.
    e_name: str
        energy_parameter_name as defined in mlip-phonon-defaults.json

    Returns
    -------
    tuple[float, float]
        A tuple containing:
        - delta parameter used for gap-fit, calculated as (es_var / num_of_descriptors).
        - Average number of triplets per atom.

    """
    at_ids = [atom.get_atomic_numbers() for atom in atoms_db]
    isolated_atom_energies = {
        atom.get_atomic_numbers()[0]: atom.info[e_name]
        for atom in atoms_db
        if "config_type" in atom.info and "IsolatedAtom" in atom.info["config_type"]
    }

    es_visol = np.array(
        [
            (atom.info[e_name] - sum([isolated_atom_energies[j] for j in at_ids[ct]]))
            / len(atom)
            for ct, atom in enumerate(atoms_db)
        ]
    )
    es_var = np.var(es_visol)
    cutoff = (
        cutoff * 2 / 3
    )  # two-thirds of the cutoff is used since two-body interactions are weak near its edge.
    descriptor_num_list = [
        compute_num_of_descriptor(atom=at, nb=2, cutoff=cutoff) / len(at)
        for at in atoms_db
    ]
    num_of_descriptors = sum(descriptor_num_list) / len(atoms_db)
    return es_var / num_of_descriptors


def compute_num_of_descriptor(atom: Atoms, nb: int, cutoff: float) -> list[float]:
    """
    Compute the number of two-body or three-body descriptors within a specified cutoff radius.

    Parameters
    ----------
    atom: ASE atoms object
        The structure to evaluate.
    nb: int
        Two-body or three-body interactions.
    cutoff: float
        Cutoff radius used to compute the dimensionality of the descriptor.

    Returns
    -------
    list[float]
        Returns a list of the number of pairs or triplets a structure is involved in.

    """
    if nb == 2:
        desc = descriptors.Descriptor(f"distance_2b add_species=T cutoff={cutoff}")
    elif nb == 3:
        desc = descriptors.Descriptor(
            f"distance_Nb order=3 add_species=T cutoff={cutoff}"
        )

    n_desc, _ = desc.sizes(atom)

    return n_desc


def run_ace(num_processes_fit: int, script_name: str) -> None:
    """
    Julia-ACE script runner.

    Parameters
    ----------
    num_processes_fit: int
        Number of threads to be used for the run.
    script_name: str
        Name of the Julia script to run.

    """
    os.environ["JULIA_NUM_THREADS"] = str(num_processes_fit)

    with (
        open("julia-ace_out.log", "w", encoding="utf-8") as file_out,
        open("julia-ace_err.log", "w", encoding="utf-8") as file_err,
    ):
        subprocess.call(["julia", script_name], stdout=file_out, stderr=file_err)


def run_gap(num_processes_fit: int, parameters) -> None:
    """
    GAP runner.

    num_processes_fit: int
        number of threads to be used for the run.

    Parameters
    ----------
        GAP fit parameters.

    """
    os.environ["OMP_NUM_THREADS"] = str(num_processes_fit)
    os.environ["OPENBLAS_NUM_THREADS"] = "1"  # blas library
    os.environ["BLIS_NUM_THREADS"] = "1"  # blas library
    os.environ["MKL_NUM_THREADS"] = "1"  # blas library
    os.environ["NETLIB_NUM_THREADS"] = "1"  # blas library

    with (
        open("std_gap_out.log", "w", encoding="utf-8") as file_std,
        open("std_gap_err.log", "w", encoding="utf-8") as file_err,
    ):
        subprocess.call(["gap_fit", *parameters], stdout=file_std, stderr=file_err)


class CustomPotential(quippy.potential.Potential):
    """A custom potential class that modifies the outputs of potentials."""

    def calculate(self, *args, **kwargs):
        """Update the atoms object with forces, energy, and virial information."""
        res = super().calculate(*args, **kwargs)
        atoms = kwargs["atoms"] if "atoms" in kwargs else args[0]
        if "forces" in self.results:
            atoms.arrays["forces"] = self.results["forces"].copy()
        if "energy" in self.results:
            atoms.info["energy"] = self.results["energy"].copy()
        if "stress" in self.results:
            atoms.info["stress"] = self.results["stress"].copy()
        return res


def _compute_gap_energy(atom, gap_control: str, gap_label: str):
    """
    Compute potential energy of a single ASE Atoms object.

    Parameters
    ----------
    atom : ase.Atoms
        The Atoms object for which to compute the potential energy.
    gap_control : str
        Control string specifying the GAP potential configuration
        (e.g., "Potential xml_label=<label>").
    gap_label : str
        The GAP XML potential file.
    """
    pot = CustomPotential(args_str=gap_control, param_filename=gap_label)
    atom.calc = pot
    atom.info["energy"] = atom.get_potential_energy()
    atom.arrays["force"] = atom.get_forces()
    atom.calc = None
    return atom


def run_ase_gap(
    num_processes_fit: int,
    data_path: str,
    xml_file: str,
    filename: str,
    glue_xml: bool = False,
) -> None:
    """
    ASE-GAP runner.

    Parameters
    ----------
    num_processes_fit: int
        number of threads to be used for the run.
    data_path:
        Path to the data file.
    filename: str
        Name of the output file.
    glue_xml: bool
        Use the glue.xml core potential instead of fitting 2b terms.
    """
    gap_control = "Potential xml_label=" + extract_gap_label(xml_file)
    atoms = ase.io.read(data_path, index=":")
    eval_worker = partial(
        _compute_gap_energy, gap_control=gap_control, gap_label=xml_file
    )
    with threadpool_limits(limits=1), mp.Pool(processes=num_processes_fit) as pool:
        atoms_eval = pool.map(eval_worker, atoms)
    ase.io.write(filename, atoms_eval, format="extxyz")


def run_nep(gpu_identifier_indices: list[int]) -> None:
    """
    NEP runner.

    Parameters
    ----------
    gpu_identifier_indices: list[int]
        Indices that identifies the GPU that NEP should be run with
    """
    env = os.environ.copy()
    env["CUDA_VISIBLE_DEVICES"] = ",".join(map(str, gpu_identifier_indices))

    with (
        open("std_nep_out.log", "w", encoding="utf-8") as file_out,
        open("std_nep_err.log", "w", encoding="utf-8") as file_err,
    ):
        subprocess.call("nep", stdout=file_out, stderr=file_err, env=env)


def run_nequip(command: str, log_prefix: str) -> None:
    """
    Nequip runner.

    Parameters
    ----------
    command: str
        The command to execute, along with its arguments.
    log_prefix: str
        Prefix for log file names, used to differentiate between different commands' logs.

    """
    with (
        open(f"{log_prefix}_out.log", "w", encoding="utf-8") as file_out,
        open(f"{log_prefix}_err.log", "w", encoding="utf-8") as file_err,
    ):
        subprocess.call(command.split(), stdout=file_out, stderr=file_err)


def run_mace(hypers: list) -> None:
    """
    MACE runner.

    Parameters
    ----------
    hypers: list
        containing all hyperparameters required for the MACE model training.

    """
    with (
        open("mace_train_out.log", "w", encoding="utf-8") as file_std,
        open("mace_train_err.log", "w", encoding="utf-8") as file_err,
    ):
        subprocess.call(["mace_run_train", *hypers], stdout=file_std, stderr=file_err)


def prepare_fit_environment(
    database_dir: Path,
    mlip_path: Path,
    glue_xml: bool,
    train_name: str = "train.extxyz",
    test_name: str = "test.extxyz",
    glue_name: str = "glue.xml",
) -> Path:
    """
    Prepare the environment for the fit.

    Parameters
    ----------
    database_dir: Path
        Path to database directory.
    mlip_path: Path
        Path to the MLIP fit run (cwd).
    glue_xml: bool
            use the glue.xml core potential instead of fitting 2b terms.
    train_name: str
        name of the training data file.
    test_name: str
        name of the test data file.
    glue_name: str
        name of the glue.xml file or path.

    Returns
    -------
    the MLIP file path.
    """
    os.makedirs(
        os.path.join(mlip_path, train_name.replace("train.extxyz", "")), exist_ok=True
    )
    if not Path(mlip_path / test_name).exists():
        shutil.copy(
            os.path.join(database_dir, test_name),
            os.path.join(mlip_path, test_name),
        )
    if not Path(mlip_path / train_name).exists():
        shutil.copy(
            os.path.join(database_dir, train_name),
            os.path.join(mlip_path, train_name),
        )
    if glue_xml:
        # TODO: might need to be fixed for remote connection
        shutil.copy(
            Path(glue_name),
            os.path.join(mlip_path, "glue.xml"),
        )

    return Path(os.path.join(mlip_path, train_name.replace("train.extxyz", "")))


def convert_xyz_to_structure(
    atoms_list: list,
    include_forces: bool = True,
    include_stresses: bool = True,
    ref_energy_name: str = "REF_energy",
    ref_force_name: str = "REF_forces",
    ref_virial_name: str = "REF_virial",
) -> tuple[list[Structure], list, list[object], list[object]]:
    """
    Convert extxyz to pymatgen Structure format.

    Parameters
    ----------
    atoms_list:
        list of atoms to be converted.
    include_forces: bool
        will include forces with the Structure object.
    include_stresses: bool
        will include stresses with the Structure object.
    ref_energy_name : str, optional
        Reference energy name.
    ref_force_name : str, optional
        Reference force name.
    ref_virial_name : str, optional
        Reference virial name.

    Returns
    -------
    tuple(pymatgen Structure object, energies, forces, stresses)

    """
    structures = []
    energies = []
    forces = []
    stresses = []
    for atoms in atoms_list:
        structure = AseAtomsAdaptor.get_structure(atoms)
        structures.append(structure)
        energies.append(atoms.info[ref_energy_name])
        if include_forces:
            forces.append(np.array(atoms.arrays[ref_force_name]).tolist())
        else:
            forces.append(np.zeros((len(structure), 3)).tolist())
        if include_stresses:
            # convert from eV to GPa
            virial = atoms.info[ref_virial_name] / atoms.get_volume()  # eV/Å^3
            stresses.append(np.array(virial * 160.2176565).tolist())  # eV/Å^3 -> GPa
        else:
            stresses.append(np.zeros((3, 3)).tolist())

    logging.info(f"Loaded {len(structures)} structures.")

    return structures, energies, forces, stresses


def write_after_distillation_data_split(
    distillation: bool,
    force_max: float,
    split_ratio: float,
    vasp_ref_name: str = "vasp_ref.extxyz",
    train_name: str = "train.extxyz",
    test_name: str = "test.extxyz",
    force_label: str = "REF_forces",
    energy_label: str = "REF_energy",
) -> None:
    """
    Write train.extxyz and test.extxyz after data distillation and split.

    Reject structures with large force components and split dataset into training and test datasets.

    Parameters
    ----------
    distillation: bool
        For using data distillation.
    force_max: float
        Maximally allowed force in the data set.
    split_ratio: float
        Parameter to divide the training set and the test set.
        A value of 0.1 means that the ratio of the training set to the test set is 9:1
    vasp_ref_name:
        name of the VASP reference data file.
    train_name:
        name of the training data file.
    test_name:
        name of the test data file.
    force_label: str
        label of the force entries.
    energy_label: str
        label of the energy entries.
    """
    # reject structures with large force components
    atoms = (
        data_distillation(vasp_ref_name, force_max, force_label)
        if distillation
        else ase.io.read(vasp_ref_name, index=":")
    )

    # split dataset into training and test datasets
    (train_structures, test_structures) = stratified_dataset_split(
        atoms=atoms, split_ratio=split_ratio, energy_label=energy_label
    )

    ase.io.write(train_name, train_structures, format="extxyz", append=True)
    ase.io.write(test_name, test_structures, format="extxyz", append=True)


def mace_convert_virial_to_stress(
    atoms: list[Atoms], ref_virial_name: str, out_file_name: str
) -> None:
    """
    Convert a virial vector into a stress tensor.

    Parameters
    ----------
    atoms: ase.atoms.Atoms
        input structures
    ref_virial_name: str
        virial label
    out_file_name: str
        name of output file
    """
    formatted_atoms = []
    for at in atoms:
        if ref_virial_name in at.info:
            at.info["REF_stress"] = -at.info[ref_virial_name] / at.get_volume()
            del at.info[ref_virial_name]
            formatted_atoms.append(at)

    write(out_file_name, formatted_atoms, format="extxyz")



"""Utility functions for data generation jobs."""

import logging

import numpy as np
from atomate2.forcefields.jobs import (
    ForceFieldRelaxMaker,
    ForceFieldStaticMaker,
)
from atomate2.vasp.jobs.phonons import PhononDisplacementMaker
from pymatgen.core import Structure
from pymatgen.transformations.advanced_transformations import (
    CubicSupercellTransformation,
)

# Configure the logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def ml_phonon_maker_preparation(
    calculator_kwargs: dict,
    relax_maker_kwargs: dict | None,
    static_maker_kwargs: dict | None,
    bulk_relax_maker: ForceFieldRelaxMaker,
    phonon_displacement_maker: ForceFieldStaticMaker,
    static_energy_maker: ForceFieldStaticMaker | None,
) -> tuple[
    ForceFieldRelaxMaker | None,
    ForceFieldStaticMaker | None,
    ForceFieldStaticMaker | None,
]:
    """
    Prepare the MLPhononMaker for the respective MLIP model.

    bulk_relax_maker: .ForceFieldRelaxMaker or None
        Maker to perform a tight relaxation on the bulk.
        Set to ``None`` to skip the
        bulk relaxation
    static_energy_maker: .ForceFieldStaticMaker or None
        Maker to perform the computation of the DFT energy on the bulk.
        Set to ``None`` to skip the
        static energy computation
    phonon_displacement_maker: .ForceFieldStaticMaker or None
        Maker used to compute the forces for a supercell.
    relax_maker_kwargs: dict
        Keyword arguments that can be passed to the RelaxMaker.
    static_maker_kwargs: dict
        Keyword arguments that can be passed to the StaticMaker.
    """
    if bulk_relax_maker is not None:
        bulk_relax_maker = bulk_relax_maker.update_kwargs(
            update={"calculator_kwargs": calculator_kwargs}
        )
        if relax_maker_kwargs is not None:
            bulk_relax_maker = bulk_relax_maker.update_kwargs(
                update={**relax_maker_kwargs}
            )

    if phonon_displacement_maker is not None:
        phonon_displacement_maker = phonon_displacement_maker.update_kwargs(
            update={"calculator_kwargs": calculator_kwargs}
        )
        if static_maker_kwargs is not None:
            phonon_displacement_maker = phonon_displacement_maker.update_kwargs(
                {**static_maker_kwargs}
            )
    if static_energy_maker is not None:
        static_energy_maker = static_energy_maker.update_kwargs(
            update={"calculator_kwargs": calculator_kwargs}
        )
        if static_maker_kwargs is not None:
            static_energy_maker = static_energy_maker.update_kwargs(
                update={**static_maker_kwargs}
            )

    return bulk_relax_maker, phonon_displacement_maker, static_energy_maker


def update_phonon_displacement_maker(
    lattice, phonon_displacement_maker
) -> PhononDisplacementMaker:
    """
    Update the phonon_displacement_maker.

    Parameters
    ----------
    lattice:
        (Average) lattice of the structure.
    phonon_displacement_maker:
        Maker used to compute the forces for a supercell.

    Returns
    -------
    Updated phonon_displacement_maker

    """
    if lattice > 10:
        density = 350 - 15 * int(round(lattice, 0))
        if lattice > 20:
            density = 50
        phonon_displacement_maker.input_set_generator.user_kpoints_settings = {
            "reciprocal_density": density
        }
    return phonon_displacement_maker


def check_supercells(
    structure_list: list[Structure],
    structure_names: list[str] | None = None,
    min_length: float = 18,
    max_length: float = 25,
    fallback_min_length: float = 10,
    min_atoms: int = 100,
    max_atoms: int = 500,
    tolerance: float = 0.1,
) -> None:
    """
    Check the supercell size.

    Prints log output regarding the structures matching the supercell requirements.

    Parameters
    ----------
    structure_list: list[Structure]
        List of pymatgen Structure object.
    structure_names: list[str]
        List of structure names.
    min_length: float
        Minimum length of the supercell that will be built.
    max_length: float
        Maximum length of the supercell that will be built.
    max_atoms: int
        Maximally allowed number of atoms in the supercell.
    min_atoms: int
        Minimum number of atoms in the supercell that shall be reached.
    fallback_min_length: float
        Fallback option for minimum length for exceptional cases
    tolerance: float
        Tolerance for min_atoms and max_atoms

    """
    structure_names = (
        [structure.composition.reduced_formula for structure in structure_list]
        if structure_names is None
        else structure_names
    )

    min_tolerance = 1 - tolerance
    max_tolerance = 1 + tolerance

    for name, structure in zip(structure_names, structure_list):
        matrix = reduce_supercell_size(
            structure,
            min_length=min_length,
            max_length=max_length,
            fallback_min_length=fallback_min_length,
            min_atoms=min_atoms,
            max_atoms=max_atoms,
        )
        supercell = structure.make_supercell(np.array(matrix).transpose())
        a, b, c = supercell.lattice.abc
        num_atoms = supercell.num_sites
        print(supercell)
        # check if supercells are in the requirements with a certain tolerance
        if (
            not (min_atoms * min_tolerance <= num_atoms <= max_atoms * max_tolerance)
            or (
                not (
                    fallback_min_length * min_tolerance
                    <= a
                    < max_length * max_tolerance
                )
            )
            or (
                not (
                    fallback_min_length * min_tolerance
                    <= b
                    < max_length * max_tolerance
                )
            )
            or (
                not (
                    fallback_min_length * min_tolerance
                    <= c
                    < max_length * max_tolerance
                )
            )
        ):
            logger.warning("You should not include structure %s \n", name)
            logger.info(
                "because the found supercell has the following lattice parameters: %f, %f, %f \n",
                a,
                b,
                c,
            )
            logger.info("and it has the following sites: %d \n", num_atoms)
            logger.info(
                "which usually leads to convergence issues during the DFT steps."
            )
        else:
            logger.info("%s has passed the supercell check. \n", name)


def reduce_supercell_size(
    structure: Structure,
    min_length: float = 18,
    max_length: float = 22,
    fallback_min_length: float = 12,
    min_atoms: int = 100,
    max_atoms: int = 500,
    step_size: float = 1,
) -> list:
    """
    Reduce phonopy supercell size.

    Parameters
    ----------
    structure: Structure
        The pymatgen Structure object.
    min_length: float
        Minimum length of the supercell that will be built.
    max_length: float
        Maximum length of the supercell that will be built.
    max_atoms: int
        Maximally allowed number of atoms in the supercell.
    min_atoms: int
        Minimum number of atoms in the supercell that shall be reached.
    fallback_min_length: float
        Fallback option for minimum length for exceptional cases.
    step_size: float
        The step_size which is used to increase the supercell.
        If allow_orthorhombic and force_90_degrees are both set to True,
        the chosen step_size will be automatically multiplied by 5 to
        prevent a too long search for the possible supercell.

    Returns
    -------
    list
        The supercell matrix as a list object.
    """
    if fallback_min_length >= min_length:
        fallback_min_length = min_length - 1
    for minimum in range(int(min_length), int(fallback_min_length), -1):
        try:
            transformation = CubicSupercellTransformation(
                min_length=minimum,
                max_length=max_length,
                min_atoms=min_atoms,
                max_atoms=max_atoms,
                step_size=step_size,
                allow_orthorhombic=True,
                force_90_degrees=True,
            )
            new_structure = transformation.apply_transformation(structure=structure)
            if min_atoms <= new_structure.num_sites <= max_atoms:
                return transformation.transformation_matrix.transpose().tolist()
        except AttributeError:
            try:
                transformation = CubicSupercellTransformation(
                    min_length=minimum,
                    max_length=max_length,
                    min_atoms=min_atoms,
                    max_atoms=max_atoms,
                    step_size=step_size,
                    allow_orthorhombic=True,
                    force_90_degrees=False,
                )
                new_structure = transformation.apply_transformation(structure=structure)
                if min_atoms <= new_structure.num_sites <= max_atoms:
                    return transformation.transformation_matrix.transpose().tolist()
            except AttributeError:
                try:
                    transformation = CubicSupercellTransformation(
                        min_length=minimum,
                        max_length=max_length,
                        min_atoms=min_atoms,
                        max_atoms=max_atoms,
                        step_size=step_size,
                    )
                    new_structure = transformation.apply_transformation(
                        structure=structure
                    )
                    if min_atoms <= new_structure.num_sites <= max_atoms:
                        return transformation.transformation_matrix.transpose().tolist()
                except AttributeError:
                    pass

    a, b, c = structure.lattice.abc
    a_factor = np.max((np.floor(max_length / a), 1))
    b_factor = np.max((np.floor(max_length / b), 1))
    c_factor = np.max((np.floor(max_length / c), 1))

    matrix = np.array([[a_factor, 0, 0], [0, b_factor, 0], [0, 0, c_factor]])
    return matrix.transpose().tolist()



"""Functions for automatic regularization and weighting of training data."""

import ast
import logging
import traceback
from contextlib import suppress

import numpy as np
from ase import Atoms
from scipy.spatial import ConvexHull

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)


# adapted from MorrowChem's RSS routines.
def set_custom_sigma(
    atoms: list[Atoms],
    reg_minmax: list[tuple],
    isolated_atom_energies: dict | None = None,
    scheme: str = "linear-hull",
    energy_name: str = "REF_energy",
    force_name: str = "REF_forces",
    virial_name: str = "REF_virial",
    element_order: list | None = None,
    max_energy: float = 20.0,
    config_type_override: dict | None = None,
    retain_existing_sigma: bool = False,
) -> list[Atoms]:
    """
    Handle automatic regularisation based on distance to convex hull, amongst other things.

    TODO: Need to make sure this works for full multi-stoichiometry systems.

    Parameters
    ----------
    atoms: (list of ase.Atoms)
        List of atoms objects to set reg. for. Usually fitting database
    reg_minmax: (list of tuples)
        List of tuples of (min, max) values for energy, force, virial sigmas
    scheme: (str)
        Method to use for regularization. Options are: linear_hull, volume-stoichiometry
        linear_hull: for single-composition system, use 2D convex hull (E, V)
        volume-stoichiometry: for multi-composition system, use 3D convex hull of (E, V, mole-fraction)
    energy_name: (str)
        Name of energy key in atoms.info
    force_name: (str)
        Name of force key in atoms.arrays
    virial_name: (str)
        Name of virial key in atoms.info
    isolated_atom_energies: (dict)
        Dictionary of isolated energies for each atomic number.
        Only needed for volume-x scheme e.g. {14: '-163.0', 8:'-75.0'}
        for SiO2
    element_order: (list)
        List of atomic numbers in order of choice (e.g. [42, 16] for MoS2)
    max_energy: (float)
        Ignore any structures with energy above hull greater than this value (eV)
    config_type_override: (dict)
        Give custom regularization for specific configuration types
    retain_existing_sigma: bool
        Whether to keep the current sigma values for specific configuration types.
        If set to True, existing sigma values for specific configurations will remain unchanged.

    e.g. reg_minmax = [(0.1, 1), (0.001, 0.1), (0.0316, 0.316), (0.0632, 0.632)]
    [(emin, emax), (semin, semax), (sfmin, sfmax), (ssvmin, svmax)]

    """
    sigs = [[], [], []]  # type: ignore
    atoms_modi = []

    if config_type_override is None:
        config_type_override = {
            "IsolatedAtom": (1e-4, 0.0, 0.0),
            "dimer": (0.1, 0.5, 0.0),
        }

    for at in atoms:
        for config_type, sigs_override in config_type_override.items():
            if at.info["config_type"] == config_type:
                at.info["energy_sigma"] = sigs_override[0]
                at.info["force_sigma"] = sigs_override[1]
                at.info["virial_sigma"] = sigs_override[2]
                atoms_modi.append(at)

        if at.info["config_type"] == "IsolatedAtom":
            at.calc = None  # TODO side-effect alert
            del at.info["force_sigma"]
            del at.info["virial_sigma"]
            continue

        if at.info["config_type"] == "dimer":
            with suppress(Exception):
                del at.info[virial_name]

            continue

    isolated_atom_energies = isolated_atom_energies or {}

    if scheme == "linear-hull":
        logging.info("Regularising with linear hull")
        hull, points = get_convex_hull(atoms, energy_name=energy_name)
        get_e_distance_func = get_e_distance_to_hull

    elif scheme == "volume-stoichiometry":
        logging.info("Regularising with 3D volume-mole fraction hull")
        if len(isolated_atom_energies) == 0:
            raise ValueError("Need to supply dictionary of isolated energies.")

        isolated_atom_energies = {
            ast.literal_eval(k) if isinstance(k, str) else k: v
            for k, v in isolated_atom_energies.items()
        }
        points = label_stoichiometry_volume(
            atoms, isolated_atom_energies, energy_name, element_order=element_order
        )  # label atoms with volume and mole fraction
        hull = calculate_hull_nd(points)
        get_e_distance_func = get_e_distance_to_hull_nd  # type: ignore

    points = {}
    for group in sorted(
        {  # check if set comprehension is running as supposed to
            at.info.get("rss_group")
            for at in atoms
            if not at.info.get("rss_nonperiodic")
        }
    ):
        points[group] = []

    for at in atoms:
        try:
            # skip non-periodic configs, volume is meaningless
            if at.info.get("rss_nonperiodic"):
                continue
            points[at.info.get("rss_group")].append(at)
        except Exception:
            pass

    for group, atoms_group in points.items():
        logging.info(f"group: {group}")

        for val in atoms_group:

            if retain_existing_sigma and "energy_sigma" in val.info:
                atoms_modi.append(val)
                continue

            de = get_e_distance_func(
                hull,
                val,
                energy_name=energy_name,
                isolated_atom_energies=isolated_atom_energies,
                element_order=element_order,
            )

            if de > max_energy:
                # don't even fit if too high
                continue
            if val.info["config_type"] not in config_type_override:
                if group == "initial":
                    sigs[0].append(reg_minmax[1][1])
                    sigs[1].append(reg_minmax[2][1])
                    sigs[2].append(reg_minmax[3][1])
                    val.info["energy_sigma"] = reg_minmax[1][1]
                    val.info["force_sigma"] = reg_minmax[2][1]
                    val.info["virial_sigma"] = reg_minmax[3][1]
                    atoms_modi.append(val)
                    continue

                if de <= reg_minmax[0][0]:
                    sigs[0].append(reg_minmax[1][0])
                    sigs[1].append(reg_minmax[2][0])
                    sigs[2].append(reg_minmax[3][0])
                    val.info["energy_sigma"] = reg_minmax[1][0]
                    val.info["force_sigma"] = reg_minmax[2][0]
                    val.info["virial_sigma"] = reg_minmax[3][0]
                    atoms_modi.append(val)

                elif de >= reg_minmax[0][1]:
                    sigs[0].append(reg_minmax[1][1])
                    sigs[1].append(reg_minmax[2][1])
                    sigs[2].append(reg_minmax[3][1])
                    val.info["energy_sigma"] = reg_minmax[1][1]
                    val.info["force_sigma"] = reg_minmax[2][1]
                    val.info["virial_sigma"] = reg_minmax[3][1]
                    atoms_modi.append(val)

                else:
                    [e, f, v] = piecewise_linear(
                        de,
                        [
                            (
                                0.1,
                                [reg_minmax[1][0], reg_minmax[2][0], reg_minmax[3][0]],
                            ),
                            (
                                1.0,
                                [reg_minmax[1][1], reg_minmax[2][1], reg_minmax[3][1]],
                            ),
                        ],
                    )
                    sigs[0].append(e)
                    sigs[1].append(f)
                    sigs[2].append(v)
                    val.info["energy_sigma"] = e
                    val.info["force_sigma"] = f
                    val.info["virial_sigma"] = v
                    atoms_modi.append(val)

    labels = ["E", "F", "V"]
    data_type = [np.array(sig) for sig in sigs]  # [e, f, v]

    for label, data in zip(labels, data_type):
        if len(data) == 0:
            logging.info(
                "No automatic regularisation performed (no structures requested)"
            )
            continue
        if label == "E":
            # Report of the regularisation statistics
            logging.info(
                f"Automatic regularisation statistics for {len(data)} structures:\n"
            )
            logging.info(
                "{:>20s}{:>20s}{:>20s}{:>20s}{:>20s}".format(
                    "", "Mean", "Std", "Nmin", "Nmax"
                )
            )
        logging.info(
            f"{label:>20s}"
            f"{data.mean():>20.4f}"
            f"{data.std():>20.4f}"
            f"{(data == data.min()).sum():>20d}"
            f"{(data == data.max()).sum():>20d}"
        )

    return atoms_modi


def get_convex_hull(
    atoms, energy_name="energy", **kwargs
) -> tuple[np.ndarray, np.ndarray]:
    """
    Calculate the simple linear (E,V) convex hull.

    Parameters
    ----------
    atoms: list
        List of atoms objects.
    energy_name: str
        Name of the energy key in atoms.info (typically a DFT energy).

    Returns
    -------
    tuple
        A tuple containing two elements:
        - lower_half_hull_points: list of points (volume, energy) in the convex hull (lower half only).
        - p: list of all points for testing purposes.

    """
    points_list = []
    failed_count = 0

    for atom in atoms:
        if atom.info["config_type"] in ["IsolatedAtom", "dimer"]:
            continue
        try:
            volume_per_atom = atom.get_volume() / len(atom)
            energy_per_atom = (
                atom.info[energy_name] / len(atom)
                if energy_name != "energy"
                else atom.get_potential_energy() / len(atom)
            )
            points_list.append((volume_per_atom, energy_per_atom))
        except KeyError:
            failed_count += 1

    if failed_count > 0:
        raise ValueError(
            f"Convex hull failed to include {failed_count}/{len(atoms)} structures"
        )

    points = np.array(points_list)
    points = points.T[:, np.argsort(points.T[0])].T  # sort by volume axis

    hull = ConvexHull(points)  # generate full convex hull
    hull_points = points[hull.vertices]

    min_x_index = np.argmin(hull_points[:, 0])
    max_x_index = np.argmax(hull_points[:, 0])

    lower_half_hull = []
    i = min_x_index

    while True:
        lower_half_hull.append(hull.vertices[i])
        i = (i + 1) % len(hull.vertices)
        if i == max_x_index:
            lower_half_hull.append(hull.vertices[i])
            break

    lower_half_hull_points = points[lower_half_hull]

    lower_half_hull_points = lower_half_hull_points[
        lower_half_hull_points[:, 1] <= np.max(lower_half_hull_points[:, 1])
    ]

    return lower_half_hull_points, points


def get_e_distance_to_hull(
    hull: np.array, atoms, energy_name="energy", **kwargs
) -> float:
    """
    Calculate the distance of a structure to the linear convex hull in energy.

    Parameters
    ----------
    hull: (np.array)
        Points in the convex hull
    atoms: (Atoms)
        Structure to calculate distance to hull
    energy_name: (str)
        Name of the energy key in atoms.info (typically a DFT energy)

    """
    volume = atoms.get_volume() / len(atoms)
    energy = (
        atoms.info[energy_name] / len(atoms)
        if energy_name != "energy"
        else atoms.get_potential_energy() / len(atoms)
    )
    tp = np.array([volume, energy])
    hull_ps = hull.points if isinstance(hull, ConvexHull) else hull

    if any(
        np.isclose(hull_ps[:], tp).all(1)
    ):  # if the point is on the hull, return 0.0
        return 0.0

    nearest = np.searchsorted(hull_ps.T[0], tp, side="right")[
        0
    ]  # find the nearest convex hull point

    return (
        energy
        - get_intersect(
            tp,  # get intersection of the vertical line (energy axis)
            tp + np.array([0, 1]),  # and the line between the nearest hull points
            hull_ps[(nearest - 1) % len(hull_ps.T[0])],
            hull_ps[nearest % len(hull_ps.T[0])],
        )[1]
    )


def get_intersect(a1, a2, b1, b2) -> tuple[float, float] | tuple:
    """
    Return the point of intersection of the lines passing through a2,a1 and b2,b1.

    a1: [x, y]
        A point on the first line
    a2: [x, y]
        Another point on the first line
    b1: [x, y]
        A point on the second line
    b2: [x, y]
        Another point on the second line

    """
    s = np.vstack([a1, a2, b1, b2])  # s for stacked
    h = np.hstack((s, np.ones((4, 1))))  # h for homogeneous
    l1 = np.cross(h[0], h[1])  # get first line
    l2 = np.cross(h[2], h[3])  # get second line
    x, y, z = np.cross(l1, l2)  # point of intersection
    if z == 0:  # lines are parallel
        return (float("inf"), float("inf"))
    return x / z, y / z


def get_mole_frac(atoms, element_order=None) -> float | int:
    """
    Calculate the mole-fraction of a structure.

    Parameters
    ----------
    atoms: (Atoms)
        Structure to calculate mole-fraction of
    element_order: (list)
        List of atomic numbers in order of choice (e.g. [42, 16] for MoS2)

    Returns
    -------
    (x2, x3...): (array of float)
        Reduced mole-fraction of structure - first element n = 1-sum(others)

    """
    element, cts = np.unique(atoms.get_atomic_numbers(), return_counts=True)

    if element_order is None and len(element) < 3:  # compatibility with old version
        x = cts[1] / sum(cts) if len(element) == 2 else 1

    else:  # new version, requires element_order, recommended for all new calculations
        if element_order is None:
            element_order = element  # use default order
        not_in = [i for i in element_order if i not in element]
        for i in not_in:
            element = np.insert(element, -1, i)
            cts = np.insert(cts, -1, 0)

        cts = np.array(
            [cts[np.argwhere(element == i).squeeze()] for i in element_order]
        )
        element = np.array(
            [element[np.argwhere(element == i).squeeze()] for i in element_order]
        )

        x = cts[1:] / sum(cts)

    return x


def label_stoichiometry_volume(
    atoms_list: list[Atoms],
    isolated_atom_energies: dict,
    energy_name: str,
    element_order: list | None = None,
) -> np.ndarray:
    """
    Calculate the stoichiometry, energy, and volume coordinates for forming the convex hull.

    Parameters
    ----------
    atoms_list: (list[Atoms])
        List of atoms objects
    isolated_atom_energies: (dict)
        Dictionary of isolated atom energies {atomic_number: energy}
    energy_name: (str)
        Name of energy key in atoms.info (typically a DFT energy)
    element_order: (list | None)
        List of atomic numbers in order of choice (e.g. [42, 16] for MoS2)

    """
    isolated_atom_energies = {
        ast.literal_eval(k) if isinstance(k, str) else k: v
        for k, v in isolated_atom_energies.items()
    }
    points_list = []
    for atom in atoms_list:
        try:
            volume = atom.get_volume() / len(atom)
            # make energy relative to isolated atoms
            energy = (
                atom.info[energy_name]
                - sum([isolated_atom_energies[j] for j in atom.get_atomic_numbers()])
                if energy_name != "energy"
                else atom.get_potential_energy()
                - sum([isolated_atom_energies[j] for j in atom.get_atomic_numbers()])
            ) / len(atom)
            mole_frac = get_mole_frac(atom, element_order=element_order)
            points_list.append(np.hstack((mole_frac, volume, energy)))
        except KeyError:
            traceback.print_exc()
    points = np.array(points_list)
    return points[np.lexsort(points.T[::-1])]


def point_in_triangle_2D(p1, p2, p3, pn) -> bool:
    """
    Check if a point is inside a triangle in 2D.

    Parameters
    ----------
    p1: (tuple)
        Coordinates of first point
    p2: (tuple)
        Coordinates of second point
    p3: (tuple)
        Coordinates of third point
    pn: (tuple)
        Coordinates of point to check

    """
    ep = 1e-4
    x1, y1 = p1
    x2, y2 = p2
    x3, y3 = p3
    x, y = pn

    denominator = (y2 - y3) * (x1 - x3) + (x3 - x2) * (y1 - y3)
    a = ((y2 - y3) * (x - x3) + (x3 - x2) * (y - y3)) / denominator
    b = ((y3 - y1) * (x - x3) + (x1 - x3) * (y - y3)) / denominator
    c = 1 - a - b

    return (
        0 - ep <= a
        and a <= 1 + ep
        and 0 - ep <= b
        and b <= 1 + ep
        and 0 - ep <= c
        and c <= 1 + ep
    )


def point_in_simplex_nd(pn, *preg) -> bool:
    """
    Check if a point is inside a region of hyperplanes in N dimensions.

    Make a little convex hull in N-1 D and check this.

    Parameters
    ----------
    pn:
        Point to check (in ND)
    *preg:
        A list of points defining an (N-1)D simplex.

    """
    preg = np.array(preg)
    pn = np.array(pn)
    if preg.shape[1] != pn.shape[0]:
        raise ValueError(
            f"Point and points must have the same dimensionality. Got {pn.shape[0]} and {preg.shape[1]}."
        )
    hull = ConvexHull(preg)
    return np.all(np.dot(hull.equations[:, :-1], pn) + hull.equations[:, -1] <= 1e-10)


def calculate_hull_nd(points_nd) -> ConvexHull:
    """
    Calculate the convex hull in ND (N>=3).

    Parameters
    ----------
    points_nd:
        Point in ND.

    Returns
    -------
    Convex hull in ND.

    """
    p0 = np.array(
        [
            (points_nd[:, i].max() - points_nd[:, i].min()) / 2 + points_nd[:, i].min()
            for i in range(points_nd.shape[1] - 1)
        ]
        + [-1e6]
    )  # test point to get the visible facets from below
    pn = np.vstack((p0, points_nd))
    remove_dim = []

    for i in range(points_nd.shape[1]):
        if np.all(points_nd.T[i, 0] == points_nd.T[i, :]):
            pn = np.delete(pn, i, axis=1)
            logging.info(f"Convex hull lower dimensional - removing dimension {i}")
            remove_dim.append(i)

    hull = ConvexHull(pn, qhull_options="QG0")
    logging.info("done calculating hull")
    hull.remove_dim = remove_dim

    return hull


def get_e_distance_to_hull_nd(
    hull, atoms, isolated_atom_energies=None, energy_name="energy", element_order=None
) -> float:
    """
    Calculate the energy distance to the N-dimensional convex hull.

    Parameters
    ----------
    hull:
        Convex hull.
    atoms: (ase.Atoms)
        Structure to calculate mole-fraction of
    isolated_atom_energies: (dict)
        Dictionary of isolated atom energies
    energy_name: (str)
        Name of energy key in atoms.info (typically a DFT energy)
    element_order: (list)
        List of atomic numbers in order of choice (e.g. [42, 16] for MoS2)

    """
    isolated_atom_energies = {
        ast.literal_eval(k) if isinstance(k, str) else k: v
        for k, v in isolated_atom_energies.items()
    }
    mole_frac = get_mole_frac(atoms, element_order=element_order)
    energy = (
        atoms.info[energy_name]
        - sum([isolated_atom_energies[j] for j in atoms.get_atomic_numbers()])
        if energy_name != "energy"
        else atoms.get_potential_energy()
        - sum([isolated_atom_energies[j] for j in atoms.get_atomic_numbers()])
    ) / len(atoms)
    volume = atoms.get_volume() / len(atoms)
    sp = np.hstack([mole_frac, volume, energy])

    for i in hull.remove_dim:
        sp = np.delete(sp, i)

    if len(sp[:-1]) == 1:
        logging.info("doing convexhull analysis in 1D")
        return get_e_distance_to_hull(hull, atoms, energy_name=energy_name)

    for _, visible_facet in enumerate(hull.simplices[hull.good]):
        if point_in_simplex_nd(sp[:-1], *hull.points[visible_facet][:, :-1]):
            n_d = hull.points[visible_facet]

            if n_d.shape[1] == 3:
                norm = np.cross(n_d[2] - n_d[0], n_d[1] - n_d[0])
                plane_norm = norm / np.linalg.norm(norm)
            else:
                relative_vectors = n_d[:-1] - n_d[0]
                plane_norm = np.linalg.lstsq(
                    relative_vectors, np.ones(relative_vectors.shape[0]), rcond=None
                )[0]

            plane_constant = np.dot(plane_norm, n_d[0])

            return (
                energy
                - (plane_constant - np.dot(plane_norm[:-1], sp[:-1])) / plane_norm[-1]
            )

    logging.info("Failed to find distance to hull in ND")
    return 1e6


def piecewise_linear(x, vals) -> np.ndarray:
    """
    Piecewise linear.

    Parameters
    ----------
    x:
        The x value.
    vals:
        The values

    """
    i = np.searchsorted([v[0] for v in vals], x)
    f0 = (vals[i][0] - x) / (vals[i][0] - vals[i - 1][0])
    return f0 * np.array(vals[i - 1][1]) + (1.0 - f0) * np.array(vals[i][1])



"""Utility functions for rss."""

import ast
import json
import os
from functools import partial
from multiprocessing import Pool
from pathlib import Path
from typing import Literal

import ase.io
import matgl
import numpy as np
from ase import Atoms
from ase.constraints import (
    FixConstraint,
    FixSymmetry,
    UnitCellFilter,
    slice2enlist,
)
from ase.data import atomic_numbers, chemical_symbols
from ase.geometry import find_mic
from ase.optimize.precon import Exp, PreconLBFGS
from ase.units import GPa
from mace.calculators import MACECalculator
from matgl.ext.ase import M3GNetCalculator
from nequip.ase import NequIPCalculator
from pymatgen.core import Structure
from pymatgen.io.ase import AseAtomsAdaptor
from threadpoolctl import threadpool_limits

from autoplex.fitting.common.utils import (
    CustomPotential,
    extract_gap_label,
)


def extract_pairstyle(
    ace_label: str, ace_json: str, ace_table: str
) -> tuple[dict[str, int], list[str]]:
    """
    Extract the pair style and coefficients from ACE potential files for running LAMMPS.

    Parameters
    ----------
        ace_label: str
            Label for the ACE potential.
        ace_json: str
            Path to the JSON file of ACE potential.
        ace_table: str
            Path to the table file containing pairwise coefficients of ACE potential.
    """
    with open(ace_json) as file:
        data = json.load(file)

    E0 = data["IP"]["components"][2]["E0"]

    elements = list(E0.keys())

    sorted_elements = sorted(elements, key=lambda x: atomic_numbers[x])

    with open(ace_table) as file:
        lines = file.readlines()

        for line in lines:
            if line.strip().startswith("N "):
                n_value = line.strip().split()[1]
                break

    elements_str = " ".join(sorted_elements)

    cmds = [
        f"pair_style     hybrid/overlay pace table spline {n_value}",
        f"pair_coeff     * * pace {ace_label} {elements_str}",
    ]

    atom_types = {}

    for i in range(len(sorted_elements)):
        atom_types[sorted_elements[i]] = i + 1

        for j in range(i, len(sorted_elements)):
            pairs = sorted_elements[i] + "_" + sorted_elements[j]
            cmds.append(f"pair_coeff     {i+1} {j+1} table {ace_table} {pairs}")

    return atom_types, cmds


class HookeanRepulsion(FixConstraint):
    """Constrain atoms softly to a minimum separation.

    Intended to avoid early iterations of potentials
    causing crashes due to overlapping atoms.

    It is recommended to calibrate the spring constant for your system
    dependent on the potential and the atomic species used. It is not guaranteed
    that the constraint will be either soft enough (e.g. non-exploding in MD) or
    strong enough (to avoid overlaps) for all spring constants and distances.

    References
    ----------
    *    Title: ASE constraints package at  at ase/ase/constraints.py
    *    Author: Ask Hjorth Larsen
    *    Date 07/10/2024
    *    Code version: 3.23.0
    *    Availability: https://gitlab.com/ase/
    """

    def __init__(
        self,
        a1: int,
        a2: (
            int
            | tuple[float, float, float]
            | tuple[float, float, float, float]
            | Literal["cell"]
        ),
        k: float,
        rt: float | None = None,
    ) -> None:
        """Apply a Hookean restorative force repel two atoms that are close.

        Parameters
        ----------
        a1: int
           Atom 1 index
        a2: can be one of three options
           1) Atom 2 index
           2) a fixed point in cartesian space to which to tether a1
           3) a plane given as (A, B, C, D) in A x + B y + C z + D = 0.
           4) 'cell' :: selects the unit cell to constrain
        k: float
           Hooke's law (spring) constant to apply when distance
           exceeds threshold_length. Units are eV Å^-2.
        rt: float
           Threshold length above which no Hookean force is applied.
           This argument is not supplied in case 3. Units of Å.

        Notes
        -----
        If a plane is specified, the Hooke's law force is applied if the atom
        is on the normal side of the plane. For instance, the plane with
        (A, B, C, D) = (0, 0, 1, -7) defines a plane in the xy plane with a z
        intercept of +7 and a normal vector pointing in the +z direction.
        If the atom has z > 7, then a downward force would be applied of
        k * (atom.z - 7). The same plane with the normal vector pointing in
        the -z direction would be given by (A, B, C, D) = (0, 0, -1, 7).

        References
        ----------
           Andrew A. Peterson,  Topics in Catalysis volume 57, pages40-53 (2014)
           https://link.springer.com/article/10.1007%2Fs11244-013-0161-8
        """
        if isinstance(a2, int):
            self._type = "two atoms"
            self.indices = [a1, a2]
        elif len(a2) == 3:
            self._type = "point"
            self.index = a1
            self.origin = np.array(a2)
        elif len(a2) == 4:
            self._type = "plane"
            self.index = a1
            self.plane = a2

        else:
            raise RuntimeError("Unknown type for a2")
        self.threshold = rt
        self.spring = k
        self.used = False

    def get_removed_dof(self, atoms):
        """Get number of removed degrees of freedom due to constraint."""
        return 0

    def todict(self):
        """Convert constraint to dictionary."""
        dct = {"name": "Hookean"}
        dct["kwargs"] = {"rt": self.threshold, "k": self.spring}
        if self._type == "two atoms":
            dct["kwargs"]["a1"] = self.indices[0]
            dct["kwargs"]["a2"] = self.indices[1]
        elif self._type == "point":
            dct["kwargs"]["a1"] = self.index
            dct["kwargs"]["a2"] = self.origin
        elif self._type == "plane":
            dct["kwargs"]["a1"] = self.index
            dct["kwargs"]["a2"] = self.plane
        else:
            raise NotImplementedError(f"Bad type: {self._type}")
        return dct

    def adjust_positions(self, atoms, newpositions):
        """Adjust positions to match the constraints.

        Do nothing for this constraint.
        """

    def adjust_momenta(self, atoms, momenta):
        """Adjust momenta to match the constraints.

        Do nothing for this constraint.
        """

    def adjust_forces(self, atoms, forces):
        """Adjust forces on the atoms to match the constraints."""
        positions = atoms.positions
        if self._type == "plane":
            A, B, C, D = self.plane
            x, y, z = positions[self.index]
            d = (A * x + B * y + C * z + D) / np.sqrt(A**2 + B**2 + C**2)
            if d < 0:
                return
            magnitude = self.spring * d
            direction = -np.array((A, B, C)) / np.linalg.norm((A, B, C))
            forces[self.index] += direction * magnitude
            return
        if self._type == "two atoms":
            p1, p2 = positions[self.indices]
        elif self._type == "point":
            p1 = positions[self.index]
            p2 = self.origin

        displace, _ = find_mic(p2 - p1, atoms.cell, atoms.pbc)
        bondlength = np.linalg.norm(displace)

        if bondlength < self.threshold:
            print(
                "Hookean adjusting forces, bondlength: ",
                bondlength,
                " < ",
                self.threshold,
            )
            self.used = True
            magnitude = self.spring * (self.threshold - bondlength)
            direction = displace / np.linalg.norm(displace)
            if self._type == "two atoms":
                forces[self.indices[0]] -= direction * magnitude
                forces[self.indices[1]] += direction * magnitude
            else:
                forces[self.index] += direction * magnitude

    def adjust_potential_energy(self, atoms):
        """Return the difference to the potential energy due to an active constraint.

        (the quantity returned is to be added to the potential energy).
        """
        positions = atoms.positions
        if self._type == "plane":
            A, B, C, D = self.plane
            x, y, z = positions[self.index]
            d = (A * x + B * y + C * z + D) / np.sqrt(A**2 + B**2 + C**2)
            if d > 0:
                return 0.5 * self.spring * d**2
            return 0.0
        if self._type == "two atoms":
            p1, p2 = positions[self.indices]
        elif self._type == "point":
            p1 = positions[self.index]
            p2 = self.origin
        displace, _ = find_mic(p2 - p1, atoms.cell, atoms.pbc)
        bondlength = np.linalg.norm(displace)
        if bondlength < self.threshold:
            return 0.5 * self.spring * (bondlength - self.threshold) ** 2
        return 0.0

    def get_indices(self):
        """Get the indices."""
        if self._type == "two atoms":
            return self.indices
        if self._type == "point":
            return self.index
        if self._type == "plane":
            return self.index
        return None

    def index_shuffle(self, atoms, ind):
        """Change the indices."""
        if self._type == "two atoms":
            newa = [-1, -1]  # Error condition
            for new, old in slice2enlist(ind, len(atoms)):
                for i, a in enumerate(self.indices):
                    if old == a:
                        newa[i] = new
            if newa[0] == -1 or newa[1] == -1:
                raise IndexError("Constraint not part of slice")
            self.indices = newa
        elif (self._type == "point") or (self._type == "plane"):
            new_a = -1  # Error condition
            for new, old in slice2enlist(ind, len(atoms)):
                if old == self.index:
                    new_a = new
                    break
            if new_a == -1:
                raise IndexError("Constraint not part of slice")
            self.index = new_a

    def __repr__(self):
        """Return a representation of the constraint."""
        if self._type == "two atoms":
            return f"Hookean({self.indices[0]}, {self.indices[1]})"
        if self._type == "point":
            return f"Hookean({self.index}) to cartesian"
        return f"Hookean({self.index}) to plane"


def process_rss(
    atom: Atoms,
    mlip_type: str,
    mlip_path: str | Path,
    output_file_name: str = "RSS_relax_results",
    scalar_pressure_method: str = "exp",
    scalar_exp_pressure: float = 100,
    scalar_pressure_exponential_width: float = 0.2,
    scalar_pressure_low: float = 0,
    scalar_pressure_high: float = 50,
    max_steps: int = 1000,
    force_tol: float = 0.01,
    stress_tol: float = 0.01,
    hookean_repul: bool = False,
    hookean_paras: dict | None = None,
    write_traj: bool = True,
    device: str = "cpu",
    isolated_atom_energies: dict[int, float] | None = None,
    config_type: str = "traj",
    keep_symmetry: bool = True,
) -> str | None:
    """Run RSS on a single thread using MLIPs.

    Parameters
    ----------
    atom: Atoms
        ASE Atoms object representing the atomic configuration.
    mlip_type: str
        Choose one specific MLIP type:
        'GAP' | 'J-ACE' | 'NequIP' | 'M3GNet' | 'MACE'.
    mlip_path: str | Path
        Path to the MLIP model.
    output_file_name: str
        Prefix for the trajectory/log file name. The actual output file name
        may be composed of this prefix, an index, and file types.
    scalar_pressure_method: str
        Method for adding external pressures. Default is 'exp'.
    scalar_exp_pressure: float
        Scalar exponential pressure. Default is 100.
    scalar_pressure_exponential_width: float
        Width for scalar pressure exponential. Default is 0.2.
    scalar_pressure_low: float
        Low limit for scalar pressure. Default is 0.
    scalar_pressure_high: float
        High limit for scalar pressure. Default is 50.
    max_steps: int
        Maximum number of steps for relaxation. Default is 1000.
    force_tol: float
        Force residual tolerance for relaxation. Default is 0.01.
    stress_tol: float
        Stress residual tolerance for relaxation. Default is 0.01.
    hookean_repul: bool
        If true, apply Hookean repulsion. Default is False.
    hookean_paras: dict[tuple[int, int], tuple[float, float]]
        Parameters for Hookean repulsion as a dictionary of tuples. Default is None.
    write_traj: bool
        If true, write trajectory of RSS. Default is True.
    device: str
        Specify the device to use: "cuda" or "cpu".
    isolated_atom_energies: dict
        Dictionary of isolated atoms energies.
    config_type: str
        Specify the type of configurations generated from RSS.
    keep_symmetry: bool
        If true, preserve symmetry during relaxation.

    Returns
    -------
    str | None
        Output string containing path for the results of the RSS relaxation.
    """
    if hookean_paras is not None:
        hookean_paras = {
            ast.literal_eval(k) if isinstance(k, str) else k: v
            for k, v in hookean_paras.items()
        }

    if mlip_type == "GAP":
        gap_label = os.path.join(mlip_path, "gap_file.xml")
        gap_control = "Potential xml_label=" + extract_gap_label(gap_label)
        pot = CustomPotential(args_str=gap_control, param_filename=gap_label)

    elif mlip_type == "J-ACE":
        from ase.calculators.lammpslib import LAMMPSlib  # noqa

        try:
            from lammps import lammps  # noqa

            lmp = lammps()
            if "ML-PACE" not in lmp.installed_packages:
                raise RuntimeError(
                    "To use RSS flow with J-ACE potential, it needs LAMMPS compiled with"
                    "lammps-user-pace library. ML-PACE is not found in the current LAMMPS binary installation."
                    "Please follow the instructions in the autoplex documentation to install"
                    "LAMMPS with the required packages. Link to the documentation: "
                    " https://autoatml.github.io/autoplex/user/index.html#lammps-installation"
                )
        except Exception as exc:
            raise RuntimeError(
                "To use RSS flow with J-ACE potential, it needs LAMMPS compiled with"
                "python bindings and lammps-user-pace. Please follow the instructions in"
                "the autoplex documentation to install LAMMPS with the required packages. "
                "Link to the documentation:"
                " https://autoatml.github.io/autoplex/user/index.html#lammps-installation"
            ) from exc

        ace_label = os.path.join(mlip_path, "acemodel.yace")
        ace_json = os.path.join(mlip_path, "acemodel.json")
        ace_table = os.path.join(mlip_path, "acemodel_pairpot.table")

        atom_types, cmds = extract_pairstyle(ace_label, ace_json, ace_table)

        pot = LAMMPSlib(
            lmpcmds=cmds, atom_types=atom_types, log_file="test.log", keep_alive=True
        )

    elif mlip_type == "NEQUIP":
        nequip_label = os.path.join(mlip_path, "deployed_nequip_model.pth")
        if isolated_atom_energies:
            ele_syms = [
                chemical_symbols[int(e_num)] for e_num in isolated_atom_energies
            ]

        else:
            raise ValueError("isol_es is empty or not defined!")
        pot = NequIPCalculator.from_deployed_model(
            model_path=nequip_label,
            device=device,
            species_to_type_name={s: s for s in ele_syms},
            set_global_options=False,
        )

    elif mlip_type == "M3GNET":
        pot_file = matgl.load_model(path=mlip_path)
        pot = M3GNetCalculator(potential=pot_file)

    elif mlip_type == "MACE":
        mace_label = os.path.join(mlip_path, "checkpoints/MACE_model_run-123.model")
        pot = MACECalculator(model_paths=mace_label, device=device)

    unique_starting_index = atom.info["unique_starting_index"]
    log_file = output_file_name + "_" + str(unique_starting_index) + ".log"
    constraint_list = []
    if hookean_repul and hookean_paras:
        atom_num = atom.get_atomic_numbers()
        for i in range(len(atom_num)):
            for j in range(i + 1, len(atom_num)):
                if (
                    (atom_num[i], atom_num[j]) in hookean_paras
                    and hookean_paras[(atom_num[i], atom_num[j])][0] != 0
                    and hookean_paras[(atom_num[i], atom_num[j])][1] != 0
                ):
                    # print(f"Hookean repulsion is used for {atom_num[i]}-{atom_num[j]}!")
                    constraint_list.append(
                        HookeanRepulsion(
                            i, j, *hookean_paras[(atom_num[i], atom_num[j])]
                        )
                    )
                elif (
                    (atom_num[j], atom_num[i]) in hookean_paras
                    and hookean_paras[(atom_num[j], atom_num[i])][0] != 0
                    and hookean_paras[(atom_num[j], atom_num[i])][1] != 0
                ):
                    # print(f"Hookean repulsion is used for {atom_num[j]}-{atom_num[i]}!")
                    constraint_list.append(
                        HookeanRepulsion(
                            i, j, *hookean_paras[(atom_num[j], atom_num[i])]
                        )
                    )

    if keep_symmetry:
        print("Creating FixSymmetry calculator and maintaining initial symmetry!")
        constraint_list.append(FixSymmetry(atom, symprec=1.0e-4))

    if constraint_list:
        atom.set_constraint(constraint_list)

    atom.calc = pot

    if scalar_pressure_method == "exp":
        scalar_pressure_tmp = scalar_exp_pressure * GPa
        if scalar_pressure_exponential_width > 0.0:
            scalar_pressure_tmp *= np.random.exponential(
                scalar_pressure_exponential_width
            )
    elif scalar_pressure_method == "uniform":
        scalar_pressure_tmp = (
            np.random.uniform(low=scalar_pressure_low, high=scalar_pressure_high) * GPa
        )
    atom.info["RSS_applied_pressure"] = scalar_pressure_tmp / GPa
    atom = UnitCellFilter(atom, scalar_pressure=scalar_pressure_tmp)

    try:
        optimizer = PreconLBFGS(
            atom, precon=Exp(3), use_armijo=True, logfile=log_file, master=True
        )
        traj = []

        def build_traj():
            atom_copy = atom.copy()
            atom_copy.info["energy"] = atom.atoms.get_potential_energy()
            atom_copy.info["enthalpy"] = atom.get_potential_energy().copy()
            traj.append(atom_copy)

        optimizer.attach(build_traj)
        optimizer.run(fmax=force_tol, smax=stress_tol, steps=max_steps)

        minim_stat = "converged" if optimizer.converged() else "unconverged"

        for traj_at_i, traj_at in enumerate(traj):
            traj_at.info["RSS_minim_iter"] = traj_at_i
            traj_at.info["config_type"] = config_type
            traj_at.info["minim_stat"] = minim_stat

        if write_traj:
            traj_file_name = (
                output_file_name + "_traj_" + str(unique_starting_index) + ".extxyz"
            )
            ase.io.write(traj_file_name, traj, parallel=False)
        del traj[-1].info["minim_stat"]
        traj[-1].info["config_type"] = minim_stat + "_minimum"

        local_minima = traj[-1]

        if local_minima.info["config_type"] == "converged_minimum":
            dir_path = Path.cwd()
            return os.path.join(dir_path, traj_file_name)
        return None

    except RuntimeError:
        print("RuntimeError occurred during optimization! Return none!")
        return None


def minimize_structures(
    mlip_type: Literal["GAP", "J-ACE", "NEP", "NEQUIP", "M3GNET", "MACE"],
    mlip_path: str | Path,
    iteration_index: str,
    structures: list[Structure],
    output_file_name: str = "RSS_relax_results",
    scalar_pressure_method: str = "exp",
    scalar_exp_pressure: float = 100,
    scalar_pressure_exponential_width: float = 0.2,
    scalar_pressure_low: float = 0,
    scalar_pressure_high: float = 50,
    max_steps: int = 1000,
    force_tol: float = 0.01,
    stress_tol: float = 0.01,
    hookean_repul: bool = False,
    hookean_paras: dict[tuple[int, int], tuple[float, float]] | None = None,
    write_traj: bool = True,
    num_processes_rss: int = 1,
    device: str = "cpu",
    isolated_atom_energies: dict[int, float] | None = None,
    config_type: str = "traj",
    struct_start_index: int = 0,
    keep_symmetry: bool = True,
) -> list[str | None]:
    """Run RSS in parallel.

    Parameters
    ----------
    mlip_type: Literal["GAP", "J-ACE", "NEP", "NEQUIP", "M3GNET", "MACE"]
        Choose one specific MLIP type to be fitted.
    mlip_path: str | Path
        Path to the MLIP model.
    iteration_index: str
        Index for the current iteration.
    structures: list[Structure]
        List of structures to be relaxed.
    output_file_name: str
        Prefix for the trajectory/log file name. The actual output file name
        may be composed of this prefix, an index, and file types.
    scalar_pressure_method: str
        Method for adding external pressures. Default is 'exp'.
    scalar_exp_pressure: float
        Scalar exponential pressure. Default is 100.
    scalar_pressure_exponential_width: float
        Width for scalar pressure exponential. Default is 0.2.
    scalar_pressure_low: float
        Low limit for scalar pressure. Default is 0.
    scalar_pressure_high: float
        High limit for scalar pressure. Default is 50.
    max_steps: int
        Maximum number of steps for relaxation. Default is 1000.
    force_tol: float
        Force residual tolerance for relaxation. Default is 0.01.
    stress_tol: float
        Stress residual tolerance for relaxation. Default is 0.01.
    hookean_repul: bool
        If true, apply Hookean repulsion. Default is False.
    hookean_paras: dict
        Parameters for Hookean repulsion as a dictionary of tuples. Default is None.
    write_traj: bool
        If true, write trajectory of RSS. Default is True.
    num_processes_rss: int
        Number of processes used for running RSS.
    device: str
        Specify device to use "cuda" or "cpu".
    isolated_atom_energies: dict
        Dictionary of isolated atoms energies.
    config_type: str
        Specify the type of configurations generated from RSS
    struct_start_index: int
        Specify the starting index within a list
    keep_symmetry: bool
        If true, preserve symmetry during relaxation.

    Returns
    -------
    list
        Output list[str] containing paths for the results of the RSS relaxation.
    """
    atoms = [AseAtomsAdaptor().get_atoms(structure) for structure in structures]
    for at in atoms:
        if "virial" in at.info:
            del at.info["virial"]

    if hookean_repul:
        print("Hookean repulsion is used!")

    for i, atom in enumerate(atoms):
        atom.info["unique_starting_index"] = iteration_index + f"{i+struct_start_index}"

    minimize_worker = partial(
        process_rss,
        mlip_type=mlip_type,
        mlip_path=mlip_path,
        output_file_name=output_file_name,
        scalar_pressure_method=scalar_pressure_method,
        scalar_exp_pressure=scalar_exp_pressure,
        scalar_pressure_exponential_width=scalar_pressure_exponential_width,
        scalar_pressure_low=scalar_pressure_low,
        scalar_pressure_high=scalar_pressure_high,
        max_steps=max_steps,
        force_tol=force_tol,
        stress_tol=stress_tol,
        hookean_repul=hookean_repul,
        hookean_paras=hookean_paras,
        write_traj=write_traj,
        device=device,
        isolated_atom_energies=isolated_atom_energies,
        config_type=config_type,
        keep_symmetry=keep_symmetry,
    )

    with threadpool_limits(limits=1), Pool(processes=num_processes_rss) as pool:
        results = pool.map(minimize_worker, atoms)

    return list(results)


def split_structure_into_groups(structures: list, num_groups: int) -> list[list]:
    """
    Split a list of structures into several groups, with each group being its own list.

    Parameters
    ----------
    structures: list
        List of structures
    num_groups: int
        Number of structure groups, used for assigning tasks across multiple nodes,
        with each node handling one group.
    """
    base_size = len(structures) // num_groups
    remainder = len(structures) % num_groups

    structure_groups = []
    start_index = 0

    for i in range(num_groups):
        extra = 1 if i < remainder else 0
        group_size = base_size + extra

        structure_groups.append(structures[start_index : start_index + group_size])
        start_index += group_size

    return structure_groups
